import typing as t

import pydantic as pyd

from pipelime.stages import SampleStage

if t.TYPE_CHECKING:
    from pipelime.sequences import Sample


class StageDuplicateKey(SampleStage, title="duplicate-key"):
    """Duplicate an item."""

    source_key: str = pyd.Field(..., description="The key to duplicate.")
    copy_to: str = pyd.Field(
        ...,
        description=(
            "The new key name. If a key with this name already exists, "
            "it will NOT be overwritten."
        ),
    )

    def __call__(self, x: "Sample") -> "Sample":
        return x.duplicate_key(self.source_key, self.copy_to)


class StageKeyFormat(SampleStage, title="format-key"):
    """Changes key names following a format string."""

    key_format: str = pyd.Field(
        "*",
        description=(
            "The new sample key format. Any `*` will be replaced with the "
            "source key, eg, `my_*_key` on [`image`, `mask`] generates "
            "`my_image_key` and `my_mask_key`. If no `*` is found, the string is "
            "suffixed to source key, ie, `MyKey` on `image` gives "
            "`imageMyKey`. If empty, the source key will not be changed."
        ),
    )
    apply_to: t.Sequence[str] = pyd.Field(
        default_factory=list,
        description=(
            "The keys to apply the new format to. Leave empty to apply to all keys."
        ),
    )

    @pyd.validator("key_format")
    def validate_key_format(cls, v):
        if "*" in v:
            return v
        return "*" + v

    def __call__(self, x: "Sample") -> "Sample":
        keys = list(x.keys())
        for k in keys:
            x = x.rename_key(k, self.key_format.replace("*", k))
        return x


class StageRemap(SampleStage, title="remap-key"):
    """Remaps keys in sample preserving internal values."""

    remap: t.Mapping[str, str] = pyd.Field(
        ..., description="`old_key: new_key` dictionary remapping."
    )
    remove_missing: bool = pyd.Field(
        True,
        description="If TRUE missing keys in remap will be removed "
        "in the output sample before name remapping",
    )

    def __call__(self, x: "Sample") -> "Sample":
        if self.remove_missing:
            x = x.extract_keys(*self.remap.keys())
        for kold, knew in self.remap.items():
            x = x.rename_key(kold, knew)
        return x


class StageKeysFilter(SampleStage, title="filter-keys"):
    """Filters sample keys."""

    key_list: t.Sequence[str] = pyd.Field(..., description="List of keys to preserve.")
    negate: bool = pyd.Field(
        False,
        description=(
            "TRUE to delete `key_list`, FALSE delete all but keys in `key_list`."
        ),
    )

    def __call__(self, x: "Sample") -> "Sample":
        return (
            x.remove_keys(*self.key_list)
            if self.negate
            else x.extract_keys(*self.key_list)
        )



import typing as t
from enum import Enum
from pathlib import Path

from pydantic import Field

from pipelime.piper import PipelimeCommand, PiperPortType


def _get_command(cmd) -> PipelimeCommand:
    from pipelime.cli.utils import PipelimeSymbolsHelper

    if isinstance(cmd, PipelimeCommand):
        return cmd

    cmd_name, cmd_args = next(iter(cmd.items()))

    cmd_cls = PipelimeSymbolsHelper.get_command(cmd_name)
    if cmd_cls is None or not issubclass(cmd_cls[1], PipelimeCommand):
        PipelimeSymbolsHelper.show_error_and_help(
            cmd_name, should_be_cmd=True, should_be_op=False, should_be_stage=False
        )
        raise ValueError(f"{cmd_name} is not a pipelime command.")
    cmd_cls = cmd_cls[1]
    return cmd_cls() if cmd_args is None else cmd_cls(**cmd_args)


class RunCommand(PipelimeCommand, title="run"):
    """Executes a DAG of pipelime commands."""

    nodes: t.Mapping[
        str, t.Union[PipelimeCommand, t.Mapping[str, t.Optional[t.Mapping[str, t.Any]]]]
    ] = Field(
        ...,
        alias="n",
        description=(
            "A DAG of commands as a `<node>: <command>` mapping. The command can be a "
            "`<name>: <args>` mapping, where `<name>` is `pipe`, `clone`, `split` etc, "
            "while `<args>` is a mapping of its arguments."
        ),
        piper_port=PiperPortType.INPUT,
    )
    include: t.Union[str, t.Sequence[str], None] = Field(
        None, description="Nodes not in this list are not run."
    )
    exclude: t.Union[str, t.Sequence[str], None] = Field(
        None, description="Nodes in this list are not run."
    )
    token: t.Optional[str] = Field(
        None,
        alias="t",
        description=(
            "The execution token. If not specified, a new token will be generated."
        ),
    )
    watch: bool = Field(
        True, alias="w", description="Monitor the execution in the current console."
    )
    successful: bool = Field(
        None,
        description="True if the execution was successful",
        exclude=True,
        repr=False,
        piper_port=PiperPortType.OUTPUT,
    )

    def run(self):
        import uuid

        from pipelime.piper.executors.factory import NodesGraphExecutorFactory
        from pipelime.piper.graph import DAGNodesGraph
        from pipelime.piper.model import DAGModel

        if not self.token:
            self.token = uuid.uuid1().hex

        inc_n = [self.include] if isinstance(self.include, str) else self.include
        exc_n = [self.exclude] if isinstance(self.exclude, str) else self.exclude

        def _node_to_run(node: str) -> bool:
            return (inc_n is None or node in inc_n) and (
                exc_n is None or node not in exc_n
            )

        nodes = {
            name: _get_command(cmd)
            for name, cmd in self.nodes.items()
            if _node_to_run(name)
        }
        dag = DAGModel(nodes=nodes)
        graph = DAGNodesGraph.build_nodes_graph(dag)
        executor = NodesGraphExecutorFactory.get_executor(watch=self.watch)
        self.successful = executor.exec(graph, token=self.token)


class DrawCommand(PipelimeCommand, title="draw"):
    """Draws a pipelime DAG."""

    class DrawBackendChoice(Enum):
        GRAPHVIZ = "graphviz"
        MERMAID = "mermaid"

    nodes: t.Mapping[
        str, t.Union[PipelimeCommand, t.Mapping[str, t.Optional[t.Mapping[str, t.Any]]]]
    ] = Field(
        ...,
        alias="n",
        description=(
            "A DAG of commands as a `<node>: <command>` mapping. The command can be a "
            "`<name>: <args>` mapping, where `<name>` is `pipe`, `clone`, `split` etc, "
            "while `<args>` is a mapping of its arguments."
        ),
        piper_port=PiperPortType.INPUT,
    )
    include: t.Union[str, t.Sequence[str], None] = Field(
        None, description="Nodes not in this list are not run."
    )
    exclude: t.Union[str, t.Sequence[str], None] = Field(
        None, description="Nodes in this list are not run."
    )
    output: t.Optional[Path] = Field(
        None,
        alias="o",
        description=(
            "The output file. If not specified, the graph will be shown in a window."
        ),
        piper_port=PiperPortType.OUTPUT,
    )
    raw_output: t.Optional[str] = Field(
        None,
        description="raw graph representation",
        exclude=True,
        repr=False,
        piper_port=PiperPortType.OUTPUT,
    )
    backend: DrawBackendChoice = Field(
        DrawBackendChoice.GRAPHVIZ, alias="b", description="The graph backend to use."
    )
    open: bool = Field(
        False,
        description=(
            "If `output` has been set, open the image file in the default viewer."
        ),
    )
    raw: bool = Field(
        False, alias="r", description="Show the raw graph representation."
    )

    def run(self):
        import os
        import platform
        import subprocess

        from pipelime.piper.drawing.factory import NodesGraphDrawerFactory
        from pipelime.piper.graph import DAGNodesGraph
        from pipelime.piper.model import DAGModel

        def start_file(filename: str):
            if platform.system() == "Darwin":  # macOS
                subprocess.call(("open", filename))
            elif platform.system() == "Windows":  # Windows
                os.startfile(filename)
            else:  # linux variants #TODO: verify!
                subprocess.call(("xdg-open", filename))

        inc_n = [self.include] if isinstance(self.include, str) else self.include
        exc_n = [self.exclude] if isinstance(self.exclude, str) else self.exclude

        def _node_to_run(node: str) -> bool:
            return (inc_n is None or node in inc_n) and (
                exc_n is None or node not in exc_n
            )

        nodes = {
            name: _get_command(cmd)
            for name, cmd in self.nodes.items()
            if _node_to_run(name)
        }
        dag = DAGModel(nodes=nodes)
        graph = DAGNodesGraph.build_nodes_graph(dag)
        drawer = NodesGraphDrawerFactory.create(self.backend.value)

        # raw graph representation
        if self.raw:
            self.raw_output = drawer.representation(graph)
        else:
            # Show or Write
            if self.output is not None:
                drawer.export(graph, str(self.output))
                if self.open:
                    start_file(str(self.output))
            else:
                from PIL import Image

                graph_image = drawer.draw(graph=graph)
                img = Image.fromarray(graph_image, "RGB")
                img.show("Graph")



import time
from typing import List

import numpy as np
from pydantic import Field

from pipelime.commands.interfaces import InputDatasetInterface, OutputDatasetInterface
from pipelime.items import YamlMetadataItem
from pipelime.piper import PipelimeCommand, PiperPortType
from pipelime.sequences import SamplesSequence


class FakeDetector(PipelimeCommand, title="fake_detector"):

    inputs: List[InputDatasetInterface] = Field(..., piper_port=PiperPortType.INPUT)
    outputs: List[OutputDatasetInterface] = Field(..., piper_port=PiperPortType.OUTPUT)
    fake_time: float = Field(0.0, description="Fake time.")

    def run(self) -> None:
        if len(self.inputs) != len(self.outputs):
            raise RuntimeError("Number of inputs and outputs must be the same.")

        for input_, output in zip(self.inputs, self.outputs):
            # Read input sequence
            seq = input_.create_reader()

            # Add metadata
            out_seq = SamplesSequence.from_list([])  # type: ignore
            for sample in self.track(seq):
                data = {"keypoints": [np.random.randint(0, 100, (10, 4)).tolist()]}
                sample["fake_detection"] = YamlMetadataItem(data)
                out_seq.append(sample)
                time.sleep(self.fake_time)

            # Write output sequence
            out_seq = output.append_writer(out_seq)
            [... for x in self.track(seq)]



from io import TextIOWrapper, BytesIO
from typing import TextIO

import pytest
from pipelime.piper.progress.model import ProgressUpdate, OperationInfo
from pipelime.piper.progress.tracker.loguru import LoguruTrackCallback

from loguru import logger


@pytest.fixture(scope="function")
def loguru_sink() -> TextIO:
    textio = TextIOWrapper(BytesIO(), encoding="utf-8")
    sinkid = logger.add(sink=textio, level="INFO")
    yield textio
    logger.remove(sinkid)


class TestLoguruTrackCallback:
    def test_on_start(self, loguru_sink: TextIO):
        # Create callback
        callback = LoguruTrackCallback()

        # Create a progress update
        op_info = OperationInfo(
            token="token", node="node", chunk=1, message="msg", total=10
        )
        progress_update = ProgressUpdate(op_info=op_info, progress=0, finished=False)

        def _check_log():
            loguru_sink.seek(0)
            assert "token" in loguru_sink.readlines()[-1]

        # Assert that the loguru sink was called on start hook
        callback.on_start(progress_update)
        _check_log()

        # Assert that the loguru sink was called on advance hook
        callback.on_advance(progress_update)
        _check_log()

        # Assert that the loguru sink was called on finish hook
        callback.on_finish(progress_update)
        _check_log()



import json
from pathlib import Path
from typing import Any

import numpy as np
from pydantic import BaseModel

from pipelime.choixe.ast.nodes import LiteralNode, Node
from pipelime.choixe.visitors.unparser import Unparser


class Decoder(Unparser):
    """Specialization of the `Unparser` for the decode operation."""

    def visit_literal(self, node: LiteralNode) -> Any:
        data = super().visit_literal(node)
        if isinstance(data, np.ndarray):
            return data.tolist()
        elif "numpy" in str(type(data)):
            return data.item()
        elif isinstance(data, BaseModel):
            symbol = f"{data.__module__}.{data.__class__.__qualname__}"
            return {"$model": symbol, "$args": json.loads(data.json(by_alias=True))}
        elif isinstance(data, Path):
            return str(data)
        else:
            return data


def decode(node: Node) -> Any:
    """A special unparsing operation that also converts some object types like numpy
    arrays into built-in lists that are supported by common markup formats like yaml and
    json.

    Args:
        node (Node): The Choixe AST node to decode.

    Returns:
        List[Tuple[List[Union[str, int]], Any]]: The decoded unparsed node.
    """
    return node.accept(Decoder())



from pipelime.sequences import Sample
from pipelime.items import UnknownItem


class TestKeyStages:
    def _check_target_pred(self, target: Sample, pred: Sample):
        assert all(
            k in target and type(target[k]) == type(v) and target[k]() == v()
            for k, v in pred.items()
        )
        assert all(
            k in pred and type(pred[k]) == type(v) and pred[k]() == v()
            for k, v in target.items()
        )

    def test_remap_keep(self):
        from pipelime.stages import StageRemap

        stage = StageRemap(remap={"a": "b"}, remove_missing=False)
        self._check_target_pred(
            Sample({"b": UnknownItem(10), "c": UnknownItem(20)}),
            stage(Sample({"a": UnknownItem(10), "c": UnknownItem(20)})),
        )

    def test_remap_remove(self):
        from pipelime.stages import StageRemap

        stage = StageRemap(remap={"a": "b"}, remove_missing=True)
        self._check_target_pred(
            Sample({"b": UnknownItem(10)}),
            stage(Sample({"a": UnknownItem(10), "c": UnknownItem(20)})),
        )

    def test_remap_keep_dont_overwrite(self):
        from pipelime.stages import StageRemap

        stage = StageRemap(remap={"a": "b"}, remove_missing=False)
        self._check_target_pred(
            Sample({"a": UnknownItem(10), "b": UnknownItem(20)}),
            stage(Sample({"a": UnknownItem(10), "b": UnknownItem(20)})),
        )

    def test_remap_remove_and_overwrite(self):
        from pipelime.stages import StageRemap

        stage = StageRemap(remap={"a": "b"}, remove_missing=True)
        self._check_target_pred(
            Sample({"b": UnknownItem(10)}),
            stage(Sample({"a": UnknownItem(10), "b": UnknownItem(20)})),
        )

    def test_filter_positive(self):
        from pipelime.stages import StageKeysFilter

        stage = StageKeysFilter(key_list=["a", "b"], negate=False)
        self._check_target_pred(
            Sample({"a": UnknownItem(10), "b": UnknownItem(20)}),
            stage(
                Sample(
                    {"a": UnknownItem(10), "b": UnknownItem(20), "c": UnknownItem(30)}
                )
            ),
        )

    def test_filter_negate(self):
        from pipelime.stages import StageKeysFilter

        stage = StageKeysFilter(key_list=["a", "b"], negate=True)
        self._check_target_pred(
            Sample({"c": UnknownItem(30)}),
            stage(
                Sample(
                    {"a": UnknownItem(10), "b": UnknownItem(20), "c": UnknownItem(30)}
                )
            ),
        )

    def test_replace(self):
        from pipelime.stages import StageReplaceItem
        from pipelime.items import NpyNumpyItem

        stage = StageReplaceItem(key_item_map={"a": NpyNumpyItem})
        self._check_target_pred(
            Sample({"a": NpyNumpyItem(10), "c": UnknownItem(20)}),
            stage(Sample({"a": UnknownItem(10), "c": UnknownItem(20)})),
        )



from abc import ABC, abstractmethod


class Estimator(ABC):
    """An estimator for the speed and ETA of a running task"""

    @abstractmethod
    def tick(self, advance: int = 1) -> None:
        """Advance the estimator by a custom amount of steps

        Args:
            advance (int, optional): The number of steps to advance. Defaults to 1.
        """
        pass

    @abstractmethod
    def reset(self, total: int) -> None:
        """Reset the estimator to a new total number of steps

        Args:
            total (int): The total number of steps
        """
        pass

    @property
    @abstractmethod
    def eta(self) -> float:
        """The estimated time remaining in seconds"""
        pass

    @property
    @abstractmethod
    def speed(self) -> float:
        """The estimated speed in steps per second"""
        pass

    @property
    @abstractmethod
    def elapsed(self) -> float:
        """The elapsed time in seconds"""
        pass

    @property
    @abstractmethod
    def start_time(self) -> float:
        """The start time in seconds"""
        pass



import os
import typing as t
from pathlib import Path

import pytest


@pytest.fixture(scope="session")
def data_folder() -> Path:
    return Path(__file__).parent / "sample_data"


@pytest.fixture(scope="session")
def datasets_folder(data_folder: Path) -> Path:
    return data_folder / "datasets"


@pytest.fixture(scope="session")
def items_folder(data_folder: Path) -> Path:
    return data_folder / "items"


@pytest.fixture(scope="session")
def augmentations_folder(data_folder: Path) -> Path:
    return data_folder / "augmentations"


@pytest.fixture(scope="session")
def choixe_folder(data_folder: Path) -> Path:
    return data_folder / "choixe"


@pytest.fixture(scope="session")
def extra_modules(data_folder: Path) -> t.List[t.Dict[str, t.Any]]:
    return [
        {
            "filepath": (data_folder / "cli" / "extra_commands.py"),
            "operators": [],
            "commands": ["randrange"],
        },
        {
            "filepath": (data_folder / "cli" / "extra_operators.py"),
            "operators": ["reversed"],
            "commands": [],
        },
    ]


@pytest.fixture(scope="session")
def minimnist_dataset(datasets_folder: Path) -> dict:
    return {
        "path": datasets_folder / "underfolder_minimnist",
        "root_keys": ["cfg", "numbers", "pose"],
        "item_keys": ["image", "label", "mask", "metadata", "points"],
        "image_keys": ["image", "mask"],
        "len": 20,
    }


@pytest.fixture(scope="function")
def minimnist_private_dataset(minimnist_dataset: dict, tmp_path: Path) -> dict:
    from shutil import copytree

    dest = Path(
        copytree(str(minimnist_dataset["path"]), str(tmp_path / "minimnist_private"))
    )
    minimnist_dataset["path"] = dest
    return minimnist_dataset


@pytest.fixture(scope="session")
def choixe_plain_cfg(choixe_folder: Path) -> Path:
    return choixe_folder / "plain_cfg.yml"


@pytest.fixture(scope="function")
def minio(tmp_path: Path):
    minio_path = os.environ.get("MINIO_APP")
    if not minio_path or not Path(minio_path).is_file():
        yield ""
        return

    try:
        import minio  # noqa: F401
    except ModuleNotFoundError:
        yield ""
        return

    from subprocess import Popen
    from time import sleep

    minio_root = tmp_path / ".minio"
    minio_root.mkdir()
    minio_proc = Popen([minio_path, "server", str(minio_root)])
    sleep(5)
    yield "minioadmin"

    # teardown
    minio_proc.terminate()



import json
import time

import pytest
import zmq

from pipelime.piper.progress.model import OperationInfo, ProgressUpdate
from pipelime.piper.progress.tracker.zmq import ZmqTrackCallback


@pytest.fixture(scope="function")
def zmq_socket() -> zmq.Socket:
    context = zmq.Context()
    socket = context.socket(zmq.SUB)
    socket.connect("tcp://localhost:5556")
    socket.subscribe(b"token")
    yield socket
    socket.close()


class TestLoguruTrackCallback:
    def test_on_start(self, zmq_socket: zmq.Socket):
        # Create callback
        callback = ZmqTrackCallback()

        # Create a progress update
        op_info = OperationInfo(
            token="token", node="node", chunk=1, message="msg", total=10
        )
        progress_update = ProgressUpdate(op_info=op_info, progress=0, finished=False)

        def _check_zmq():
            t0 = time.time()
            while time.time() - t0 < 10:
                try:
                    token, msg = zmq_socket.recv_multipart(flags=zmq.NOBLOCK)
                    prog = ProgressUpdate.parse_obj(json.loads(msg.decode("utf-8")))
                    assert token.decode() == "token"
                    assert prog == progress_update
                    return
                except zmq.ZMQError:
                    pass
            assert False

        # Assert that the loguru sink was called on start hook
        callback.on_start(progress_update)
        _check_zmq()

        # Assert that the loguru sink was called on advance hook
        callback.on_advance(progress_update)
        _check_zmq()

        # Assert that the loguru sink was called on finish hook
        callback.on_finish(progress_update)
        _check_zmq()



import numpy as np
import warnings
import typing as t

from pipelime.items import Item


class NumpyItem(Item[np.ndarray]):
    @classmethod
    def validate(cls, raw_data: t.Any) -> np.ndarray:
        return np.array(raw_data)


class NpyNumpyItem(NumpyItem):
    @classmethod
    def file_extensions(cls) -> t.Sequence[str]:
        return (".npy",)

    @classmethod
    def decode(cls, fp: t.BinaryIO) -> np.ndarray:
        return np.load(fp)

    @classmethod
    def encode(cls, value: np.ndarray, fp: t.BinaryIO):
        np.save(fp, value)


class TxtNumpyItem(NumpyItem):
    @classmethod
    def file_extensions(cls) -> t.Sequence[str]:
        return (".txt",)

    @classmethod
    def decode(cls, fp: t.BinaryIO) -> np.ndarray:
        with warnings.catch_warnings():
            warnings.filterwarnings(action="ignore", message="loadtxt")
            return np.atleast_1d(np.loadtxt(fp))

    @classmethod
    def encode(cls, value: np.ndarray, fp: t.BinaryIO):
        np.savetxt(fp, value)

    @classmethod
    def validate(cls, raw_data: t.Any) -> np.ndarray:
        return np.atleast_1d(np.array(raw_data))



import typing as t
import pydantic as pyd

import pipelime.commands.interfaces as pl_interfaces
from pipelime.piper import PipelimeCommand, PiperPortType
import pipelime.utils.pydantic_types as pl_types


class TimeItCommand(PipelimeCommand, title="timeit"):
    """Measures the average time to get a sample from a sequence."""

    class OutputTime(pyd.BaseModel):
        nanosec: int

        def __repr__(self) -> str:
            return self.__piper_repr__()

        def __piper_repr__(self) -> str:
            from pipelime.cli.utils import time_to_str

            return time_to_str(self.nanosec)

    input: t.Optional[
        pl_interfaces.InputDatasetInterface
    ] = pl_interfaces.InputDatasetInterface.pyd_field(
        alias="i",
        is_required=False,
        description=(
            "The input dataset. If None, the first operation "
            "must be a sequence generator."
        ),
        piper_port=PiperPortType.INPUT,
    )

    output: t.Optional[
        pl_interfaces.OutputDatasetInterface
    ] = pl_interfaces.OutputDatasetInterface.pyd_field(
        alias="o", is_required=False, piper_port=PiperPortType.OUTPUT
    )

    operations: t.Optional[pl_types.YamlInput] = pyd.Field(
        None,
        alias="op",
        description=(
            "An optional pipeline to run or a path to a yaml/json file as "
            "<filepath>[:<key-path>]\n"
            "The pipeline is defined as a mapping or a sequence of mappings where "
            "each key is a samples sequence operator to run, eg, `map`, `sort`, etc., "
            "while the value gathers the arguments, ie, a single value, a sequence of "
            "values or a keyword mapping."
        ),
    )

    skip_first: pyd.NonNegativeInt = pyd.Field(
        1, alias="s", description="Skip the first n samples."
    )
    max_samples: t.Optional[pyd.PositiveInt] = pyd.Field(
        None,
        alias="m",
        description="Grab at most `max_samples` and take the average time.",
    )
    repeat: pyd.NonNegativeInt = pyd.Field(
        0, alias="r", description="Repeat the measurement `repeat` times."
    )
    process: bool = pyd.Field(
        False,
        alias="p",
        description=(
            "Measure process time instead of using a performance counter clock."
        ),
    )

    average_time: t.Optional[OutputTime] = pyd.Field(
        None,
        description="The average time to get a sample from the sequence.",
        exclude=True,
        repr=False,
        piper_port=PiperPortType.OUTPUT,
    )

    def run(self):
        import time
        from pipelime.sequences import build_pipe, SamplesSequence

        if self.input is None and self.operations is None:
            raise ValueError("No input dataset or operation defined.")

        clock_fn = time.process_time_ns if self.process else time.perf_counter_ns

        elapsed_times = []
        for r in range(self.repeat + 1):
            seq = SamplesSequence if self.input is None else self.input.create_reader()
            if self.operations is not None:
                seq = build_pipe(self.operations.value, seq)  # type: ignore

            assert isinstance(seq, SamplesSequence)
            if self.output is not None:
                seq = self.output.append_writer(seq)

            seqit = iter(seq)
            for s in range(self.skip_first):
                _ = next(seqit)

            if self.max_samples is None:
                start = clock_fn()
                for _ in seqit:
                    pass
                end = clock_fn()
            else:
                start = clock_fn()
                for s in range(self.max_samples):
                    _ = next(seqit)
                end = clock_fn()
            elapsed_times.append(end - start)

        self.average_time = TimeItCommand.OutputTime(
            nanosec=int(sum(elapsed_times) // len(elapsed_times))
        )


class PipeCommand(PipelimeCommand, title="pipe"):
    """A general-purpose command to build up linear pipelines."""

    operations: pl_types.YamlInput = pyd.Field(
        ...,
        alias="op",
        description=(
            "The pipeline to run or a path to a yaml/json file as "
            "<filepath>[:<key-path>]\n"
            "The pipeline is defined as a mapping or a sequence of mappings where "
            "each key is a samples sequence operator to run, eg, `map`, `sort`, etc., "
            "while the value gathers the arguments, ie, a single value, a sequence of "
            "values or a keyword mapping."
        ),
    )

    input: t.Optional[
        pl_interfaces.InputDatasetInterface
    ] = pl_interfaces.InputDatasetInterface.pyd_field(
        alias="i",
        is_required=False,
        description=(
            "The input dataset. If None, the first operation "
            "must be a sequence generator."
        ),
        piper_port=PiperPortType.INPUT,
    )

    output: pl_interfaces.OutputDatasetInterface = (
        pl_interfaces.OutputDatasetInterface.pyd_field(
            alias="o", piper_port=PiperPortType.OUTPUT
        )
    )

    grabber: pl_interfaces.GrabberInterface = pl_interfaces.GrabberInterface.pyd_field(
        alias="g"
    )

    def __init__(self, **data):
        super().__init__(**data)
        if not self.operations:
            raise ValueError(f"Invalid pipeline: {self.operations}")

    def run(self):
        from pipelime.sequences import build_pipe, SamplesSequence

        seq = SamplesSequence if self.input is None else self.input.create_reader()
        seq = build_pipe(self.operations.value, seq)  # type: ignore
        seq = self.output.append_writer(seq)

        self.grabber.grab_all(
            seq,
            grab_context_manager=self.output.serialization_cm(),
            keep_order=False,
            parent_cmd=self,
            track_message=f"Writing results ({len(seq)} samples)",
        )


class CloneCommand(PipelimeCommand, title="clone"):
    """Clone a dataset. You can use this command to create a local copy of a dataset
    hosted on a remote data lake by disabling the `REMOTE_FILE` serialization option."""

    input: pl_interfaces.InputDatasetInterface = (
        pl_interfaces.InputDatasetInterface.pyd_field(
            alias="i", piper_port=PiperPortType.INPUT
        )
    )

    output: pl_interfaces.OutputDatasetInterface = (
        pl_interfaces.OutputDatasetInterface.pyd_field(
            alias="o", piper_port=PiperPortType.OUTPUT
        )
    )

    grabber: pl_interfaces.GrabberInterface = pl_interfaces.GrabberInterface.pyd_field(
        alias="g"
    )

    def run(self):
        seq = self.input.create_reader()
        seq = self.output.append_writer(seq)
        self.grabber.grab_all(
            seq,
            grab_context_manager=self.output.serialization_cm(),
            keep_order=False,
            parent_cmd=self,
            track_message=f"Cloning data ({len(seq)} samples)",
        )


class ConcatCommand(PipelimeCommand, title="cat"):
    """Concatenate two or more datasets."""

    inputs: t.Union[
        pl_interfaces.InputDatasetInterface,
        t.Sequence[pl_interfaces.InputDatasetInterface],
    ] = pl_interfaces.InputDatasetInterface.pyd_field(
        alias="i", piper_port=PiperPortType.INPUT
    )

    output: pl_interfaces.OutputDatasetInterface = (
        pl_interfaces.OutputDatasetInterface.pyd_field(
            alias="o", piper_port=PiperPortType.OUTPUT
        )
    )

    grabber: pl_interfaces.GrabberInterface = pl_interfaces.GrabberInterface.pyd_field(
        alias="g"
    )

    def run(self):
        inputs = self.inputs if isinstance(self.inputs, t.Sequence) else [self.inputs]
        input_it = iter(inputs)
        seq = next(input_it).create_reader()
        for input_ in input_it:
            seq = seq.cat(input_.create_reader())
        self.grabber.grab_all(
            self.output.append_writer(seq),
            grab_context_manager=self.output.serialization_cm(),
            keep_order=False,
            parent_cmd=self,
            track_message=f"Writing data ({len(seq)} samples)",
        )


class AddRemoteCommand(PipelimeCommand, title="remote-add"):
    """Upload samples to one or more remotes.
    Slicing options filter the samples to upload,
    but the whole dataset is always written out.
    """

    input: pl_interfaces.InputDatasetInterface = (
        pl_interfaces.InputDatasetInterface.pyd_field(
            alias="i", piper_port=PiperPortType.INPUT
        )
    )

    remotes: t.Union[
        pl_interfaces.RemoteInterface, t.Sequence[pl_interfaces.RemoteInterface]
    ] = pl_interfaces.RemoteInterface.pyd_field(alias="r")

    keys: t.Union[str, t.Sequence[str]] = pyd.Field(
        default_factory=list,
        alias="k",
        description="Keys to upload. Leave empty to upload all the keys.",
    )

    start: t.Optional[int] = pyd.Field(
        None, description="The first sample (included), defaults to the first element."
    )
    stop: t.Optional[int] = pyd.Field(
        None, description="The last sample (excluded), defaults to the whole sequence."
    )
    step: t.Optional[int] = pyd.Field(
        None, description="The slice step, defaults to 1."
    )

    output: t.Optional[
        pl_interfaces.OutputDatasetInterface
    ] = pl_interfaces.OutputDatasetInterface.pyd_field(
        alias="o",
        is_required=False,
        description="Optional output dataset with remote items.",
        piper_port=PiperPortType.OUTPUT,
    )

    grabber: pl_interfaces.GrabberInterface = pl_interfaces.GrabberInterface.pyd_field(
        alias="g"
    )

    @pyd.validator("remotes", "keys")
    def validate_remotes(cls, v):
        return (
            v if not isinstance(v, (str, bytes)) and isinstance(v, t.Sequence) else [v]
        )

    def run(self):
        from pipelime.stages import StageUploadToRemote

        original = self.input.create_reader()
        seq = original.slice(start=self.start, stop=self.stop, step=self.step).map(
            StageUploadToRemote(
                remotes=[r.get_url() for r in self.remotes],  # type: ignore
                keys_to_upload=self.keys,
            )
        )

        if self.output is None:
            self._grab_all(seq, None)
        else:
            # NB: we should always write the whole dataset,
            # even if we are uploading only a slice
            if self.start is None and self.stop is None and self.step is None:
                self._grab_all(
                    self.output.append_writer(seq), self.output.serialization_cm()
                )
            else:
                self._grab_all(seq, None)
                self._grab_all(
                    self.output.append_writer(original), self.output.serialization_cm()
                )

    def _grab_all(self, seq, sm):
        self.grabber.grab_all(
            seq,
            grab_context_manager=sm,
            keep_order=False,
            parent_cmd=self,
            track_message=f"Uploading data ({len(seq)} samples)",
        )


class RemoveRemoteCommand(PipelimeCommand, title="remote-remove"):
    """Remove one or more remote from a dataset.
    NB: data is not removed from the remote data lake."""

    input: pl_interfaces.InputDatasetInterface = (
        pl_interfaces.InputDatasetInterface.pyd_field(
            alias="i", piper_port=PiperPortType.INPUT
        )
    )

    remotes: t.Union[
        pl_interfaces.RemoteInterface, t.Sequence[pl_interfaces.RemoteInterface]
    ] = pl_interfaces.RemoteInterface.pyd_field(alias="r")

    keys: t.Union[str, t.Sequence[str]] = pyd.Field(
        default_factory=list,
        alias="k",
        description=(
            "Remove remotes on these keys only. Leave empty to affect all the keys."
        ),
    )

    output: pl_interfaces.OutputDatasetInterface = (
        pl_interfaces.OutputDatasetInterface.pyd_field(
            alias="o", piper_port=PiperPortType.OUTPUT
        )
    )

    grabber: pl_interfaces.GrabberInterface = pl_interfaces.GrabberInterface.pyd_field(
        alias="g"
    )

    @pyd.validator("remotes", "keys")
    def validate_remotes(cls, v):
        return (
            v if not isinstance(v, (str, bytes)) and isinstance(v, t.Sequence) else [v]
        )

    def run(self):
        from pipelime.stages import StageForgetSource

        remotes = [r.get_url() for r in self.remotes]  # type: ignore
        remove_all = remotes if not self.keys else []
        remove_by_key = {k: remotes for k in self.keys}

        seq = self.input.create_reader().map(
            StageForgetSource(*remove_all, **remove_by_key)
        )
        self.grabber.grab_all(
            self.output.append_writer(seq),
            grab_context_manager=self.output.serialization_cm(),
            keep_order=False,
            parent_cmd=self,
            track_message=f"Removing remotes ({len(seq)} samples)",
        )


class ValidateCommand(PipelimeCommand, title="validate"):
    """Outputs a minimal schema which will validate the given input."""

    class OutputSchemaDefinition(pyd.BaseModel):
        schema_def: t.Any

        def __repr__(self) -> str:
            return self.__piper_repr__()

        def __piper_repr__(self) -> str:
            import yaml

            return yaml.safe_dump(self.schema_def, sort_keys=False)

    class OutputCmdLineSchema(pyd.BaseModel):
        schema_def: t.Any

        def __repr__(self) -> str:
            return self.__piper_repr__()

        def __piper_repr__(self) -> str:
            return " ".join(self._flatten_dict(self.schema_def))

        def _flatten_dict(self, dict_, parent_key="", sep=".", prefix="--"):
            cmd_line = []
            for k, v in dict_.items():
                new_key = f"{parent_key}{sep}{k}" if parent_key else f"{prefix}{k}"
                if isinstance(v, t.Mapping):
                    cmd_line.extend(
                        self._flatten_dict(
                            v, parent_key=new_key, sep=sep, prefix=prefix
                        )
                    )
                elif isinstance(v, t.Sequence) and not isinstance(v, (str, bytes)):
                    cmd_line.extend(
                        self._flatten_list(
                            v, parent_key=new_key, sep=sep, prefix=prefix
                        )
                    )
                else:
                    cmd_line.append(new_key)
                    cmd_line.append(str(v))
            return cmd_line

        def _flatten_list(self, list_, parent_key, sep=".", prefix="--"):
            cmd_line = []
            for i, v in enumerate(list_):
                new_key = f"{parent_key}[{i}]"
                if isinstance(v, t.Mapping):
                    cmd_line.extend(
                        self._flatten_dict(
                            v, parent_key=new_key, sep=sep, prefix=prefix
                        )
                    )
                elif isinstance(v, t.Sequence) and not isinstance(v, (str, bytes)):
                    cmd_line.extend(
                        self._flatten_list(
                            v, parent_key=new_key, sep=sep, prefix=prefix
                        )
                    )
                else:
                    cmd_line.append(new_key)
                    cmd_line.append(str(v))
            return cmd_line

    input: pl_interfaces.InputDatasetInterface = (
        pl_interfaces.InputDatasetInterface.pyd_field(
            alias="i", piper_port=PiperPortType.INPUT
        )
    )

    max_samples: int = pyd.Field(
        0,
        alias="m",
        description=(
            "Max number of samples to consider when creating the schema.  "
            "Set to 0 to check all the samples."
        ),
    )
    root_key_path: str = pyd.Field(
        "input.schema", alias="r", description="Root key path for the output schema."
    )

    grabber: pl_interfaces.GrabberInterface = pl_interfaces.GrabberInterface.pyd_field(
        alias="g"
    )

    output_schema_def: t.Optional[OutputSchemaDefinition] = pyd.Field(
        None,
        description="yaml schema definition",
        exclude=True,
        repr=False,
        piper_port=PiperPortType.OUTPUT,
    )
    output_cmd_line_schema: t.Optional[OutputCmdLineSchema] = pyd.Field(
        None,
        description="Schema definition on command line",
        exclude=True,
        repr=False,
        piper_port=PiperPortType.OUTPUT,
    )

    def run(self):
        import json
        from pipelime.stages import StageItemInfo

        seq = self.input.create_reader()
        if self.max_samples != 0:
            seq = seq[0 : self.max_samples]  # noqa
        item_info = StageItemInfo()

        self.grabber.grab_all(
            seq,
            keep_order=False,
            parent_cmd=self,
            track_message=f"Reading data ({len(seq)} samples)",
            sample_fn=item_info,
        )

        sample_schema = {
            k: pl_interfaces.ItemValidationModel(
                class_path=info.item_type,
                is_optional=(info.count_ != len(seq)),
                is_shared=info.is_shared,
            ).dict(by_alias=True)
            for k, info in item_info.items_info().items()
        }

        sample_validation = pl_interfaces.SampleValidationInterface(
            sample_schema=sample_schema,
            ignore_extra_keys=False,
            lazy=(self.max_samples == 0),
            max_samples=self.max_samples,
        ).dict()

        def _type2str(data):
            import inspect
            from pipelime.items import Item
            from pipelime.utils.pydantic_types import ItemType

            for k, v in data.items():
                if inspect.isclass(v) and issubclass(v, Item):
                    data[k] = str(ItemType(__root__=v))
                elif isinstance(v, t.Mapping):
                    _type2str(v)
            return data

        sample_validation = _type2str(sample_validation)

        if self.root_key_path:
            import pydash as py_

            tmp_dict = {}
            py_.set_(tmp_dict, self.root_key_path, sample_validation)
            sample_validation = tmp_dict

        self.output_schema_def = ValidateCommand.OutputSchemaDefinition(
            schema_def=sample_validation
        )
        self.output_cmd_line_schema = ValidateCommand.OutputCmdLineSchema(
            schema_def=sample_validation
        )


class MapCommand(PipelimeCommand, title="map"):
    """Apply a stage on a dataset."""

    stage: t.Union[str, t.Mapping[str, t.Mapping[str, t.Any]]] = pyd.Field(
        ...,
        alias="s",
        description=(
            "A stage to apply. Can be a stage name/class_path (with no arguments) or "
            "a dictionary with the stage name/class_path as key and the arguments "
            "passed by keywords."
        ),
    )

    input: pl_interfaces.InputDatasetInterface = (
        pl_interfaces.InputDatasetInterface.pyd_field(
            alias="i", piper_port=PiperPortType.INPUT
        )
    )

    output: pl_interfaces.OutputDatasetInterface = (
        pl_interfaces.OutputDatasetInterface.pyd_field(
            alias="o", piper_port=PiperPortType.OUTPUT
        )
    )

    grabber: pl_interfaces.GrabberInterface = pl_interfaces.GrabberInterface.pyd_field(
        alias="g"
    )

    def run(self):
        seq = self.input.create_reader()
        seq = seq.map(
            self.stage if isinstance(self.stage, t.Mapping) else {self.stage: {}}
        )
        self.grabber.grab_all(
            self.output.append_writer(seq),
            grab_context_manager=self.output.serialization_cm(),
            keep_order=False,
            parent_cmd=self,
            track_message=f"Mapping data ({len(seq)} samples)",
        )


class SortCommand(PipelimeCommand, title="sort"):
    """Sort a dataset based on metadata values."""

    key_path: str = pyd.Field(
        ...,
        alias="k",
        description=(
            "A pydash-like key path. The path is built by splitting the mapping "
            "keys by `.` and enclosing list indexes within `[]`. "
            "Use `\\` to escape the `.` character."
        ),
    )

    input: pl_interfaces.InputDatasetInterface = (
        pl_interfaces.InputDatasetInterface.pyd_field(
            alias="i", piper_port=PiperPortType.INPUT
        )
    )

    output: pl_interfaces.OutputDatasetInterface = (
        pl_interfaces.OutputDatasetInterface.pyd_field(
            alias="o", piper_port=PiperPortType.OUTPUT
        )
    )

    grabber: pl_interfaces.GrabberInterface = pl_interfaces.GrabberInterface.pyd_field(
        alias="g"
    )

    def run(self):
        def _sort_key_fn(x):
            return x.deep_get(self.key_path)

        seq = self.input.create_reader()
        seq = seq.sort(_sort_key_fn)
        self.grabber.grab_all(
            self.output.append_writer(seq),
            grab_context_manager=self.output.serialization_cm(),
            keep_order=False,
            parent_cmd=self,
            track_message=f"Sorting data ({len(seq)} samples)",
        )



from __future__ import annotations
import typing as t
from pathlib import Path
import pydantic as pyd

from pipelime.items import Item


yaml_any_type = t.Union[None, str, int, float, bool, t.Mapping[str, t.Any], t.Sequence]


class YamlInput(pyd.BaseModel, extra="forbid", copy_on_model_validation="none"):
    """General yaml/json data (str, number, mapping, list...) optionally loaded from
    a yaml/json file, possibly with key path (format <filepath>[:<key>])."""

    __root__: yaml_any_type

    @property
    def value(self):
        return self.__root__

    def __str__(self) -> str:
        return str(self.__root__)

    def __repr__(self) -> str:
        return self.__piper_repr__()

    def __piper_repr__(self) -> str:
        return repr(self.__root__)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        if isinstance(value, YamlInput):
            return value
        if isinstance(value, (str, bytes)):
            value = str(value)
            filepath, _, root_key = (
                value.rpartition(":") if ":" in value else (value, None, None)
            )
            filepath = Path(filepath)
            if filepath.exists():
                import yaml
                import pydash as py_

                with filepath.open() as f:
                    value = yaml.safe_load(f)
                    if root_key is not None:
                        value = py_.get(value, root_key, default=None)
            return YamlInput(__root__=value)  # type: ignore
        if cls._check_any_type(value):
            return YamlInput(__root__=value)
        raise ValueError(f"Invalid yaml data input: {value}")

    # @classmethod
    # def _check_mapping(cls, value):
    #     for k, v in value.items():
    #         if not isinstance(k, str):
    #             return False
    #         if not cls._check_any_type(v):
    #             return False
    #     return True
    #
    # @classmethod
    # def _check_sequence(cls, value):
    #     for v in value:
    #         if not cls._check_any_type(v):
    #             return False
    #     return True

    @classmethod
    def _check_any_type(cls, value):
        if isinstance(value, (str, int, float, bool, t.Sequence)) or value is None:
            return True
        if isinstance(value, t.Mapping):
            return all(isinstance(k, str) for k in value)


def _item_type_to_string(item_type: t.Type[Item]) -> str:
    return (
        item_type.__name__
        if item_type.__module__.startswith("pipelime.items")
        else item_type.__module__ + "." + item_type.__qualname__
    )


class ItemType(pyd.BaseModel):
    """Item type definition."""

    __root__: t.Type[Item]

    class Config:
        extra = "forbid"
        copy_on_model_validation = "none"
        json_encoders = {Item: _item_type_to_string}

    @classmethod
    def make_default(cls, itype: t.Any) -> ItemType:
        return cls.validate(itype)

    @property
    def itype(self) -> t.Type[Item]:
        return self.__root__

    def __call__(self, *args, **kwargs) -> Item:
        return self.__root__(*args, **kwargs)

    def __str__(self) -> str:
        return _item_type_to_string(self.__root__)

    def __repr__(self) -> str:
        return self.__piper_repr__()

    def __piper_repr__(self) -> str:
        return repr(self.__root__)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        import inspect
        from pipelime.choixe.utils.imports import import_symbol

        if isinstance(value, ItemType):
            return value
        if inspect.isclass(value) and issubclass(value, Item):
            return ItemType(__root__=value)
        if isinstance(value, (str, bytes)):
            value = str(value)
            if "." not in value:
                value = "pipelime.items." + value
            return ItemType(__root__=import_symbol(value))
        raise ValueError(f"Invalid item type: {value}")



import pytest
import pipelime.sequences as pls
import typing as t


class TestSamplesSequences:
    def test_is_normalized(self, minimnist_dataset: dict):
        import pipelime.items as pli

        sseq = pls.SamplesSequence.from_underfolder(  # type: ignore
            folder=minimnist_dataset["path"]
        ).cat(
            pls.SamplesSequence.from_list(  # type: ignore
                [pls.Sample({"__": pli.UnknownItem()})]
            )
        )

        assert sseq.is_normalized(1)
        assert sseq.is_normalized(5)
        assert not sseq.is_normalized()

    @pytest.mark.parametrize("merge_root_items", [True, False])
    def test_from_underfolder(self, minimnist_dataset: dict, merge_root_items):
        sseq = pls.SamplesSequence.from_underfolder(  # type: ignore
            folder=minimnist_dataset["path"], merge_root_items=merge_root_items
        )

        assert isinstance(sseq, pls.SamplesSequence)
        assert minimnist_dataset["len"] == len(sseq)
        assert isinstance(sseq[0], pls.Sample)

        root_sample = sseq.root_sample  # type: ignore
        assert isinstance(root_sample, pls.Sample)

        sample_list = sseq._samples  # type: ignore
        assert isinstance(sample_list, t.Sequence)
        assert minimnist_dataset["len"] == len(sample_list)
        assert isinstance(sample_list[0], pls.Sample)

        for sample, raw_sample in zip(sseq, sample_list):
            for k, v in raw_sample.items():
                assert sample[k] is v
            for k in minimnist_dataset["root_keys"]:
                assert k in root_sample
                assert root_sample[k].is_shared
                if merge_root_items:
                    assert k in sample
                    assert sample[k] is root_sample[k]
                else:
                    assert k not in sample
            for k in minimnist_dataset["item_keys"]:
                assert k in sample
                assert not sample[k].is_shared

    def test_from_underfolder_must_exist(self):
        sseq = pls.SamplesSequence.from_underfolder(  # type: ignore
            folder="no-path", must_exist=False
        )
        assert str(sseq.folder) == "no-path"  # type: ignore
        assert not sseq.must_exist  # type: ignore

        with pytest.raises(ValueError):
            sseq = pls.SamplesSequence.from_underfolder(  # type: ignore
                folder="no-path", must_exist=True
            )

    def test_from_list(self, minimnist_dataset: dict):
        source = pls.SamplesSequence.from_underfolder(  # type: ignore
            folder=minimnist_dataset["path"], merge_root_items=False
        )
        sseq = pls.SamplesSequence.from_list([s for s in source])  # type: ignore

        assert len(source) == len(sseq)
        assert all(s1 is s2 for s1, s2 in zip(source, sseq))

    def test_to_pipe(self):
        from pipelime.stages import StageIdentity
        from pathlib import Path

        a = (
            pls.SamplesSequence.from_underfolder(  # type: ignore
                folder="no-path", must_exist=False
            )
            .map(StageIdentity())
            .slice(start=10)
        )
        b = pls.SamplesSequence.from_underfolder(  # type: ignore
            folder="no-path-2", must_exist=False
        ).shuffle()
        a = a.cat(b)

        expected_pipe = [
            {
                "from_underfolder": {
                    "folder": Path("no-path"),
                    "merge_root_items": True,
                    "must_exist": False,
                    "watch": False,
                }
            },
            {"map": {"stage": StageIdentity()}},
            {"slice": {"start": 10, "stop": None, "step": None}},
            {"cat": {"to_cat": b}},
        ]

        assert a.to_pipe(recursive=False, objs_to_str=False) == expected_pipe

        expected_pipe[3]["cat"]["to_cat"] = [
            {
                "from_underfolder": {
                    "folder": Path("no-path-2"),
                    "merge_root_items": True,
                    "must_exist": False,
                    "watch": False,
                }
            },
            {"shuffle": {"seed": None}},
        ]

        assert a.to_pipe(recursive=True, objs_to_str=False) == expected_pipe

    def test_build_pipe(self):
        from pipelime.stages import StageIdentity

        input_pipe = [
            {
                "from_underfolder": {
                    "folder": "no-path",
                    "merge_root_items": True,
                    "must_exist": False,
                }
            },
            {
                "slice": {"start": 10, "stop": None, "step": None},
                "map": {"stage": {"identity": None}},
            },
        ]

        expected_seq = (
            pls.SamplesSequence.from_underfolder(  # type: ignore
                folder="no-path", merge_root_items=True, must_exist=False
            )
            .slice(start=10)
            .map(StageIdentity())
        )

        assert pls.build_pipe(input_pipe).dict() == expected_seq.dict()

        input_pipe = {
            "from_underfolder": {
                "folder": "no-path",
                "merge_root_items": True,
                "must_exist": False,
            },
            "slice": {"start": 10, "stop": None, "step": None},
            "map": {"stage": StageIdentity()},
        }

        assert pls.build_pipe(input_pipe).dict() == expected_seq.dict()

        with pytest.raises(TypeError):
            pls.build_pipe("shuffle")



from __future__ import annotations

import typing as t

from pydantic import BaseModel, Field, PrivateAttr

from pipelime.sequences import Sample, SamplesSequence


def _build_op(
    src: t.Union[SamplesSequence, t.Type[SamplesSequence]],
    ops: t.Union[str, t.Mapping[str, t.Any]],
) -> t.Union[SamplesSequence, t.Type[SamplesSequence]]:
    def _op_call(
        seq: t.Union[SamplesSequence, t.Type[SamplesSequence]],
        name: str,
        args: t.Union[t.Mapping[str, t.Any], t.Sequence],
    ) -> SamplesSequence:
        if ":" in name:
            from pipelime.choixe.utils.imports import import_module

            module_name, _, op_name = name.rpartition(":")
            import_module(module_name)
        else:
            op_name = name

        fn = getattr(seq, op_name)
        return fn(**args) if isinstance(args, t.Mapping) else fn(*args)

    if isinstance(ops, str):
        return _op_call(src, ops, {})

    for func_name, func_args in ops.items():
        # safe wrapping
        if isinstance(func_args, (str, bytes)) or not isinstance(
            func_args, (t.Sequence, t.Mapping)
        ):
            func_args = [func_args]
        src = _op_call(src, func_name, func_args)

    return src


def build_pipe(
    pipe_list: t.Union[
        str, t.Mapping[str, t.Any], t.Sequence[t.Union[str, t.Mapping[str, t.Any]]]
    ],
    source: t.Union[SamplesSequence, t.Type[SamplesSequence]] = SamplesSequence,
) -> SamplesSequence:
    """Build a pipeline from a list of operations.

    :param pipe_list: a single sequence operator or a mapping or a sequence of mappings
    :type pipe_list: t.Union[str, t.Mapping[str, t.Any],
        t.Sequence[t.Union[str, t.Mapping[str, t.Any]]]]
    :param source: the source symbol to start with, defaults to SamplesSequence
    :type source: t.Union[SamplesSequence, t.Type[SamplesSequence]], optional
    :return: the pipeline
    :rtype: SamplesSequence
    """
    for op_item in (
        pipe_list
        if isinstance(pipe_list, t.Sequence) and not isinstance(pipe_list, str)
        else [pipe_list]
    ):
        source = _build_op(source, op_item)
    return (
        source
        if isinstance(source, SamplesSequence)
        else SamplesSequence.from_list([])  # type: ignore
    )


class DataStream(
    t.Sequence[Sample],
    BaseModel,
    extra="forbid",
    copy_on_model_validation="none",
    underscore_attrs_are_private=True,
):
    """A stream of samples, comprising an input sequence to the data and an output
    pipe to further process the samples when ready.
    """

    class _SampleGen:
        _curr_keys: t.Optional[t.Sequence[str]]
        _curr_sample: Sample

        def __call__(self, _: int) -> Sample:
            x = self._curr_sample
            if self._curr_keys:
                x = x.extract_keys(*self._curr_keys)
            return x

    input_sequence: t.Optional[SamplesSequence] = Field(
        None, description="Input data sequence."
    )
    output_pipe: t.Union[
        str, t.Mapping[str, t.Any], t.Sequence[t.Union[str, t.Mapping[str, t.Any]]]
    ] = Field(..., description="Output data pipe.")

    _output_sequence: SamplesSequence
    _sample_gen: _SampleGen = PrivateAttr(default_factory=_SampleGen)

    def __init__(self, **data):
        super().__init__(**data)
        self._output_sequence = build_pipe(
            self.output_pipe,
            SamplesSequence.from_callable(  # type: ignore
                generator_fn=self._sample_gen, length=1
            ),
        )

    @classmethod
    def read_write_underfolder(
        cls, path: str, must_exists: bool = True, zfill: t.Optional[int] = None
    ) -> DataStream:
        """Creates a DataStream to read and write samples from the same underfolder."""
        seq: SamplesSequence = SamplesSequence.from_underfolder(  # type: ignore
            path, must_exist=must_exists, watch=True
        )
        return cls(
            input_sequence=seq,  # type: ignore
            output_pipe={
                "to_underfolder": {
                    "folder": path,
                    "exists_ok": True,
                    "zfill": seq.best_zfill()
                    if zfill is None
                    else zfill,  # to be consistent if samples are added
                }
            },
        )

    @classmethod
    def create_new_underfolder(cls, path: str, zfill: int = 0) -> DataStream:
        """Creates a DataStream to write samples to a new underfolder dataset."""
        seq: SamplesSequence = SamplesSequence.from_underfolder(  # type: ignore
            path, must_exist=False, watch=True
        )
        return cls(
            input_sequence=seq,  # type: ignore
            output_pipe={
                "to_underfolder": {
                    "folder": path,
                    "exists_ok": False,
                    "zfill": zfill,
                }
            },
        )

    def __len__(self) -> int:
        return 0 if self.input_sequence is None else len(self.input_sequence)

    def __getitem__(self, idx: int) -> Sample:
        if self.input_sequence is None:
            raise IndexError("No input sequence.")

        x = self.input_sequence[idx]
        for v in x.values():
            v.cache_data = False  # always watch for changes
        return x

    def get_input(self, idx: int) -> Sample:
        return self[idx]

    def set_output(
        self,
        idx: int,
        sample: Sample,
        keys: t.Optional[t.Sequence[str]] = None,
    ):
        self._sample_gen._curr_keys = keys
        self._sample_gen._curr_sample = sample
        _ = self._output_sequence[idx]



import inspect
import typing as t

from pydantic import BaseModel
from pydantic.typing import display_as_type
from rich import box
from rich.markup import escape
from rich import get_console
from rich import print as rprint
from rich.pretty import Pretty
from rich.table import Table, Column


def _input_icon():
    return "📥"


def _output_icon():
    return "📦"


def _parameter_icon():
    return "📐"


def _short_line():
    return "━━━━━"


def print_with_style(
    val,
    style: t.Optional[str] = None,
    *,
    pretty: bool = False,
    end: str = "\n",
    indent_guides: bool = True,
):
    get_console().print(
        Pretty(val, indent_guides=indent_guides, expand_all=True) if pretty else val,
        style=style,
        end=end,
    )


def print_debug(
    val, *, pretty: bool = False, end: str = "\n", indent_guides: bool = True
):
    get_console().print(
        Pretty(val, indent_guides=indent_guides, expand_all=True) if pretty else val,
        style="italic grey50",
        end=end,
    )


def print_info(
    val, *, pretty: bool = False, end: str = "\n", indent_guides: bool = True
):
    get_console().print(
        Pretty(val, indent_guides=indent_guides, expand_all=True) if pretty else val,
        style="cyan",
        end=end,
    )


def print_warning(
    val, *, pretty: bool = False, end: str = "\n", indent_guides: bool = True
):
    get_console().print(
        "[bold blink]WARNING:[/bold blink]",
        Pretty(val, indent_guides=indent_guides, expand_all=True) if pretty else val,
        style="orange1",
        end=end,
    )


def print_error(
    val, *, pretty: bool = False, end: str = "\n", indent_guides: bool = True
):
    get_console().print(
        "[bold blink]ERROR:[/bold blink]",
        Pretty(val, indent_guides=indent_guides, expand_all=True) if pretty else val,
        style="dark_red on white",
        end=end,
    )


def get_model_title(model_cls: t.Type[BaseModel]) -> str:
    if model_cls.__config__.title:
        return model_cls.__config__.title
    return model_cls.__name__


def get_model_classpath(model_cls: t.Type[BaseModel]) -> str:
    if hasattr(model_cls, "_classpath") and model_cls._classpath:  # type: ignore
        return model_cls._classpath  # type: ignore
    return f"{model_cls.__module__}.{model_cls.__name__}"


def print_model_field_values(
    model_fields: t.Mapping,
    port_values: t.Mapping[str, t.Any],
    icon: str = "",
):
    for k, v in port_values.items():
        rprint(f"\n{icon if icon else '***'} {k}:")
        if model_fields[k].field_info.description:
            rprint(
                f"[italic grey50]{escape(model_fields[k].field_info.description)}[/]"
            )
        rprint(
            "[green]"
            + escape(str(v) if isinstance(v, (bytes, str)) else repr(v))
            + "[/]"
        )


def print_command_inputs(command: "PipelimeCommand"):  # type: ignore # noqa: E602,F821
    print_model_field_values(command.__fields__, command.get_inputs(), _input_icon())


def print_command_outputs(command: "PipelimeCommand"):  # type: ignore # noqa: E602,F821
    print_model_field_values(command.__fields__, command.get_outputs(), _output_icon())


def print_models_short_help(
    *model_cls: t.Type[BaseModel],
    show_class_path: bool = True,
):
    grid = Table.grid(
        *([Column(overflow="fold") for _ in range(2 + int(show_class_path))]),
        padding=(0, 1),
    )
    for m in model_cls:
        col_vals = [escape(get_model_title(m))]
        if show_class_path:
            col_vals.append(f"[italic grey50]{escape(get_model_classpath(m))}[/]")
        col_vals.append(escape(inspect.getdoc(m) or ""))
        grid.add_row(*col_vals)
    rprint(grid)


def print_model_info(
    model_cls: t.Type[BaseModel],
    *,
    indent_offs: int = 2,
    show_class_path: bool = True,
    show_piper_port: bool = True,
):
    grid = Table(
        *(
            [
                Column("Fields", overflow="fold"),
                Column("Description", overflow="fold"),
                Column("Type", overflow="fold"),
            ]
            + ([Column("Piper Port", overflow="fold")] if show_piper_port else [])
            + [
                Column("Default", overflow="fold"),
            ]
        ),
        box=box.SIMPLE_HEAVY,
        title=(
            f"[bold dark_red]{escape(get_model_title(model_cls))}[/]\n"
            f"[blue]{escape(_get_signature(model_cls))}[/]\n"
            f"[italic grey23]{escape(inspect.getdoc(model_cls) or '')}[/]"
        ),
        caption=escape(get_model_classpath(model_cls)) if show_class_path else None,
        title_style="on white",
        expand=True,
    )

    _iterate_model_fields(
        model_cls=model_cls,
        grid=grid,
        indent=0,
        indent_offs=indent_offs,
        show_piper_port=show_piper_port,
        add_blank_row=True,
    )

    rprint(grid)


def _field_row(
    grid: Table, field, indent: int, indent_offs: int, show_piper_port: bool
):
    is_model = _is_model(field.outer_type_) and not inspect.isabstract(
        field.outer_type_
    )
    has_root_item = ("__root__" in field.outer_type_.__fields__) if is_model else False
    field_outer_type = field.outer_type_

    if has_root_item:
        field_outer_type = field.outer_type_.__fields__["__root__"].outer_type_

    if show_piper_port:
        from pipelime.piper import PiperPortType

        fport = str(
            field.field_info.extra.get("piper_port", PiperPortType.PARAMETER).value
        ).upper()

        if fport == PiperPortType.INPUT.value.upper():
            fport = f"{_input_icon()} [yellow]{fport}[/]"
        elif fport == PiperPortType.OUTPUT.value.upper():
            fport = f"{_output_icon()} [cyan]{fport}[/]"
        else:
            fport = f"{_parameter_icon()} {fport}"
    else:
        fport = None

    line = (
        [
            (" " * indent)
            + ("[bold salmon1]" if indent == 0 else "")
            + (
                f"{escape(field.name)} / "
                if field.model_config.allow_population_by_field_name and field.has_alias
                else ""
            )
            + f"{escape(field.alias)}"
            + ("[/]" if indent == 0 else ""),
            ("▶ " + escape(field.field_info.description))
            if field.field_info.description
            else "",
            (
                ""
                if is_model and not has_root_item
                else display_as_type(field_outer_type).replace("[", r"\[")  # noqa: W605
            ),
        ]
        + ([fport] if fport else [])
        + (["[red]✗[/]"] if field.required else [f"[green]{field.get_default()}[/]"])
    )

    grid.add_row(*line)

    if is_model and not has_root_item:
        _iterate_model_fields(
            model_cls=field_outer_type,
            grid=grid,
            indent=indent + indent_offs,
            indent_offs=indent_offs,
            show_piper_port=show_piper_port,
            add_blank_row=False,
        )
    else:
        inner_types = _get_inner_args(field_outer_type)
        last_types = inner_types
        while last_types:
            last_types = _get_inner_args(*last_types)
            inner_types |= last_types
        inner_types = {
            arg for arg in inner_types if _is_model(arg) and arg is not BaseModel
        }

        for arg in inner_types:
            grid.add_row((" " * indent) + f"[grey50]{_short_line()} {arg.__name__}[/]")
            _iterate_model_fields(
                model_cls=arg,
                grid=grid,
                indent=indent + indent_offs,
                indent_offs=indent_offs,
                show_piper_port=show_piper_port,
                add_blank_row=False,
            )


def _get_signature(model_cls: t.Type[BaseModel]) -> str:
    excluded = [
        mfield.alias
        for mfield in model_cls.__fields__.values()
        if mfield.field_info.exclude
    ]
    sig = inspect.signature(model_cls)
    sig = sig.replace(
        parameters=[p for p in sig.parameters.values() if p.name not in excluded],
        return_annotation=inspect.Signature.empty,
    )
    return str(sig)


def _is_model(type_):
    return inspect.isclass(type_) and issubclass(type_, BaseModel)


def _recursive_args_flattening(arg):
    if isinstance(arg, t.Mapping):
        return {
            a
            for k, v in arg.items()
            for u in (_recursive_args_flattening(k), _recursive_args_flattening(v))
            for a in u
        }
    if isinstance(arg, t.Collection):
        return {a for v in arg for a in _recursive_args_flattening(v)}
    return {arg}


def _get_inner_args(*type_):
    return {
        a
        for t_ in type_
        for arg in t.get_args(t_)
        for a in _recursive_args_flattening(arg)
    }


def _iterate_model_fields(
    model_cls, grid, indent, indent_offs, show_piper_port, add_blank_row
):
    no_data = True
    for field in model_cls.__fields__.values():  # type: ignore
        if not field.field_info.exclude:
            no_data = False
            _field_row(
                grid,
                field,
                indent=indent,
                indent_offs=indent_offs,
                show_piper_port=show_piper_port,
            )
            if add_blank_row:
                grid.add_row()
    if no_data:
        grid.add_row((" " * indent) + "[grey50 italic]([strike]no parameters[/])[/]")



import time
from pipelime.piper.progress.estimator.base import Estimator
from pipelime.piper.progress.estimator.naive import NaiveEstimator


class TestNaiveEstimator:
    def test_estimator(self):
        N = 10
        estimator = NaiveEstimator(alpha=0.8)
        assert isinstance(estimator, Estimator)

        # Reset the estimator
        estimator.reset(total=N)

        # Check that the estimator is reset
        assert estimator.start_time > 0
        assert estimator.elapsed == 0
        assert estimator.speed < 0
        assert estimator.eta < 0

        for _ in range(N):
            # Tick the estimator
            time.sleep(0.1)
            estimator.tick(advance=1)

            # Check that the estimator is updated
            assert estimator.start_time > 0
            assert estimator.elapsed > 0
            assert estimator.speed > 0
            assert estimator.eta >= 0



import time

from pipelime.piper.progress.listener.callbacks.rich_table import (
    RichTableListenerCallback,
)
from pipelime.piper.progress.model import OperationInfo, ProgressUpdate

# This test will not perform any asserts, it will just call the methods and ensure
# no exception is thrown. TODO: Find a way to capture rich live object printed from
# another thread.


class TestRichTableListenerCallback:
    def test_callback(self):
        callback = RichTableListenerCallback()
        op_infos = [
            OperationInfo(
                token="token", node=f"node{i}", chunk=0, total=4, message="message"
            )
            for i in range(3)
        ]

        try:
            # Start the callback
            callback.on_start()

            time.sleep(0.1)

            # Update the callback
            for op_info in op_infos:
                for i in range(op_info.total):
                    callback.on_update(ProgressUpdate(op_info=op_info, progress=i))

        finally:
            # Stop the callback
            callback.on_stop()



from pipelime.piper.progress.listener.factory import (
    ProgressReceiverFactory,
    ListenerCallbackFactory,
)
from pipelime.piper.progress.listener.base import ProgressReceiver, ListenerCallback


class TestReceiverFactory:
    def test_get_receiver(self):
        receiver = ProgressReceiverFactory.get_receiver("token")
        assert isinstance(receiver, ProgressReceiver)


class TestListenerCallbackFactory:
    def test_get_callback(self):
        callback = ListenerCallbackFactory.get_callback()
        assert isinstance(callback, ListenerCallback)



from abc import ABC, abstractmethod

from loguru import logger

from pipelime.piper.graph import DAGNodesGraph
from pipelime.piper.progress.listener.base import Listener
from pipelime.piper.progress.listener.factory import (
    ListenerCallbackFactory,
    ProgressReceiverFactory,
)


class NodesGraphExecutor(ABC):
    """An NodesGraphExecutor is an abstract class that should
    implement the execution of a NodesGraph made of Nodes.
    """

    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def exec(self, graph: DAGNodesGraph, token: str = "") -> bool:
        """Executes the given NodesGraph.

        Args:
            graph (DAGNodesGraph): input DAGNodesGraph
            token (str, optional): execution token shared among nodes. Defaults to "".

        Returns:
            bool: TRUE if the execution was successful, FALSE otherwise
        """
        raise NotImplementedError()


class WatcherNodesGraphExecutor(NodesGraphExecutor):
    """Decorator for `NodesGraphExecutor` that wraps an existing executor.

    The `WatcherNodesGraphExecutor` disables all logs from the wrapped executor and
    lets a `Watcher` print live updates in the form of a rich table.

    This wrapper can be used with any subtype of `NodesGraphExecutor`::

        # Different types of executor
        executor1 = NaiveNodesGraphExecutor(token)
        executor2 = AdvancedNodesGraphExecutor(token)
        executor3 = SuperCoolNodesGraphExecutor(token)

        # All of them can be transformed into a `WatcherNodesGraphExecutor`
        # regardless of their type.
        executor1 = WatcherNodesGraphExecutor(executor1)
        executor2 = WatcherNodesGraphExecutor(executor2)
        executor3 = WatcherNodesGraphExecutor(executor3)
    """

    def __init__(self, executor: NodesGraphExecutor) -> None:
        self._executor = executor

    def exec(self, graph: DAGNodesGraph, token: str = "") -> bool:
        res = False
        logger.disable(self._executor.__module__)
        receiver = ProgressReceiverFactory.get_receiver(token)
        callback = ListenerCallbackFactory.get_callback()
        listener = Listener(receiver, callback)
        try:
            listener.start()
            res = self._executor.exec(graph, token)
        finally:
            listener.stop()
            logger.enable(self._executor.__module__)
        return res



import pytest
import pipelime.remotes as plr
import pipelime.sequences as pls

from pathlib import Path
from urllib.parse import ParseResult
from filecmp import cmp
import typing as t


class TestRemotes:
    def _upload_download(
        self,
        temp_folder,
        source: str,
        remote_url: ParseResult,
    ):
        from pipelime.remotes.base import BaseRemote

        remote = plr.create_remote(remote_url)
        assert isinstance(remote, BaseRemote)

        remote_base_path, _ = plr.paths_from_url(remote_url)
        assert remote_base_path is not None

        # upload
        source_to_remote: t.Mapping[Path, ParseResult] = {}
        reader = pls.SamplesSequence.from_underfolder(  # type: ignore
            source, merge_root_items=False
        )
        for sample in reader:
            for k, itm in sample.items():
                rm_url = remote.upload_file(itm._file_sources[0], remote_base_path)
                assert rm_url is not None
                source_to_remote[itm._file_sources[0]] = rm_url

        # download and compare
        local_root = temp_folder / "local"
        local_root.mkdir(parents=True)

        if remote_base_path.startswith("/"):
            remote_base_path = remote_base_path[1:]

        for original, rm_url in source_to_remote.items():
            rm, (rm_base_path, rm_name) = plr.create_remote(rm_url), plr.paths_from_url(
                rm_url
            )
            assert isinstance(rm, type(remote))
            assert isinstance(rm_base_path, str)
            assert isinstance(rm_name, str)
            assert rm_base_path == remote_base_path

            local_file = local_root / rm_name
            assert rm.download_file(local_file, rm_base_path, rm_name)
            assert cmp(str(local_file), original, shallow=False)

    def test_shared_folder_remote(self, minimnist_dataset: dict, tmp_path: Path):
        remote_url = plr.make_remote_url(
            scheme="file", host="localhost", bucket=(tmp_path / "rmbucket")
        )

        self._upload_download(tmp_path, minimnist_dataset["path"], remote_url)

    def test_s3_remote(self, minimnist_dataset: dict, tmp_path: Path, minio: str):
        if not minio:
            pytest.skip("MinIO unavailable")

        remote_url = plr.make_remote_url(
            scheme="s3",
            user=f"{minio}",
            password=f"{minio}",
            host="localhost",
            port=9000,
            bucket="rmbucket",
            secure=False,
        )

        self._upload_download(tmp_path, minimnist_dataset["path"], remote_url)



from __future__ import annotations
from abc import ABCMeta, abstractmethod
from pathlib import Path
from urllib.parse import ParseResult, urlparse
from enum import IntEnum
from contextlib import ContextDecorator, contextmanager
from loguru import logger
import io
import shutil
import os
import json

import typing as t

import pipelime.remotes as plr


class SerializationMode(IntEnum):
    """Standard resolution is REMOTE FILE -> HARD LINK -> FILE COPY -> NEW FILE
    or SYM LINK -> FILE COPY -> NEW FILE. You can alter this behaviour by setting
    default and disabled serialization modes."""

    CREATE_NEW_FILE = 0
    DEEP_COPY = 1
    SYM_LINK = 2
    HARD_LINK = 3
    REMOTE_FILE = 4


class ItemFactory(ABCMeta):
    """Item classes register themselves in this factory, so that they can be
    instantiated on request when loading file from disk. To store user-created data a
    specific Item should be explicitly created.
    """

    ITEM_CLASSES: t.Dict[str, t.Type["Item"]] = {}
    REMOTE_FILE_EXT = ".remote"
    ITEM_DATA_CACHE_MODE: t.Dict[t.Type["Item"], t.Optional[bool]] = {}
    ITEM_SERIALIZATION_MODE: t.Dict[t.Type["Item"], SerializationMode] = {}
    ITEM_DISABLED_SERIALIZATION_MODES: t.Dict[
        t.Type["Item"], t.Set[SerializationMode]
    ] = {}

    def __init__(cls, name, bases, dct):
        """Registers item class extensions."""
        for ext in cls.file_extensions():  # type: ignore
            if ext == cls.REMOTE_FILE_EXT:
                raise ValueError(f"{cls.REMOTE_FILE_EXT} file extension is reserved")
            cls.ITEM_CLASSES[ext] = cls  # type: ignore
        cls.ITEM_DATA_CACHE_MODE[cls] = None  # type: ignore
        cls.ITEM_SERIALIZATION_MODE[cls] = SerializationMode.REMOTE_FILE  # type: ignore
        cls.ITEM_DISABLED_SERIALIZATION_MODES[cls] = set()  # type: ignore
        super().__init__(name, bases, dct)

    @classmethod
    def get_instance(
        cls, filepath: t.Union[Path, str], shared_item: bool = False
    ) -> Item:
        """Returns the item that can handle the given file."""
        filepath = Path(filepath)
        path_or_urls = [filepath]
        ext = filepath.suffix
        if ext == cls.REMOTE_FILE_EXT:
            with filepath.open("r") as fp:
                url_list = json.load(fp)
                url_list = [urlparse(u) for u in url_list if u]
            if not url_list:
                raise ValueError(f"The file {filepath} does not contain any remote.")
            ext = Path(url_list[0].path).suffix
            path_or_urls = url_list

        item_cls = cls.ITEM_CLASSES.get(ext, UnknownItem)
        return item_cls(*path_or_urls, shared=shared_item)

    @classmethod
    def set_data_cache_mode(
        cls, item_cls: t.Type["Item"], enable_data_cache: t.Optional[bool]
    ):
        cls.ITEM_DATA_CACHE_MODE[item_cls] = enable_data_cache

    @classmethod
    def is_cache_enabled(cls, item_cls: t.Type["Item"]) -> bool:
        for base_cls in item_cls.mro():
            if issubclass(base_cls, Item):
                value = cls.ITEM_DATA_CACHE_MODE.get(base_cls, None)
                if value is not None:
                    return value
        return True

    @classmethod
    def set_serialization_mode(cls, item_cls: t.Type["Item"], mode: SerializationMode):
        cls.ITEM_SERIALIZATION_MODE[item_cls] = mode

    @classmethod
    def get_serialization_mode(cls, item_cls: t.Type["Item"]) -> SerializationMode:
        smode = cls.ITEM_SERIALIZATION_MODE[item_cls]
        for base_cls in item_cls.mro():
            if issubclass(base_cls, Item):
                other_smode = cls.ITEM_SERIALIZATION_MODE[base_cls]
                if other_smode < smode:
                    smode = other_smode
        return smode

    @classmethod
    def set_disabled_serialization_modes(
        cls,
        item_cls: t.Type["Item"],
        modes: t.Union[t.Set[SerializationMode], t.List[SerializationMode]],
    ):
        cls.ITEM_DISABLED_SERIALIZATION_MODES[item_cls] = set(modes)

    @classmethod
    def get_disabled_serialization_modes(
        cls, item_cls: t.Type["Item"]
    ) -> t.List[SerializationMode]:
        smodes = set()
        for base_cls in item_cls.mro():
            if issubclass(base_cls, Item):
                smodes |= cls.ITEM_DISABLED_SERIALIZATION_MODES[base_cls]
        return list(smodes)


def set_item_serialization_mode(mode: SerializationMode, *item_cls: t.Type["Item"]):
    """Sets serialization mode for some or all items.
    Applies to all items if no item class is given."""
    for itc in ItemFactory.ITEM_SERIALIZATION_MODE.keys() if not item_cls else item_cls:
        ItemFactory.set_serialization_mode(itc, mode)


def set_item_disabled_serialization_modes(
    modes: t.List[SerializationMode], *item_cls: t.Type["Item"]
):
    """Disables serialization modes on selected item classes.
    Applies to all items if no item class is given."""
    for itc in ItemFactory.ITEM_SERIALIZATION_MODE.keys() if not item_cls else item_cls:
        ItemFactory.set_disabled_serialization_modes(itc, modes)


def disable_item_data_cache(*item_cls: t.Type["Item"]):
    """Disables data cache on selected item classes.
    Applies to all items if no item class is given."""
    for itc in item_cls if item_cls else ItemFactory.ITEM_DATA_CACHE_MODE.keys():
        ItemFactory.set_data_cache_mode(itc, False)


def enable_item_data_cache(*item_cls: t.Type["Item"]):
    """Enables data cache on selected item classes.
    Applies to all items if no item class is given."""
    for itc in item_cls if item_cls else ItemFactory.ITEM_DATA_CACHE_MODE.keys():
        ItemFactory.set_data_cache_mode(itc, True)


class item_serialization_mode(ContextDecorator):
    """Use this class as context manager or function decorator to temporarily change
    the items' serialization mode.

    .. code-block::
       :caption: Example

       # set the serialization mode for all items
       with item_serialization_mode("HARD_LINK"):
           ...

       # set the serialization mode only for ImageItem and NumpyItem
       with item_serialization_mode(SerializationMode.HARD_LINK, ImageItem, NumpyItem):
           ...

       # apply at function invocation
       @item_serialization_mode(SerializationMode.HARD_LINK, ImageItem)
       def my_fn():
           ...
    """

    def __init__(
        self, smode: t.Union[str, SerializationMode], *item_cls: t.Type["Item"]
    ):
        self._target_mode = (
            smode if isinstance(smode, SerializationMode) else SerializationMode[smode]
        )
        self._items = set(
            item_cls if item_cls else ItemFactory.ITEM_SERIALIZATION_MODE.keys()
        )

    def __enter__(self):
        self._prev_mode = {
            itc: ItemFactory.ITEM_SERIALIZATION_MODE[itc] for itc in self._items
        }
        set_item_serialization_mode(self._target_mode, *self._items)

    def __exit__(self, exc_type, exc_value, traceback):
        for itc, val in self._prev_mode.items():
            ItemFactory.set_serialization_mode(itc, val)


class item_disabled_serialization_modes(ContextDecorator):
    """Use this class as context manager or function decorator to temporarily change
    the items' disabled serialization modes.

    .. code-block::
       :caption: Example

       # disabled serialization modes for all items
       with item_disabled_serialization_modes(["HARD_LINK", "DEEP_COPY"]):
           ...

       # disabled serialization modes only for ImageItem and NumpyItem
       with item_disabled_serialization_modes(
           SerializationMode.HARD_LINK, ImageItem, NumpyItem
       ):
           ...

       # apply at function invocation
       @item_disabled_serialization_modes(
           ["REMOTE_FILE", SerializationMode.SYM_LINK], ImageItem
       )
       def my_fn():
           ...
    """

    def __init__(
        self,
        smodes: t.Union[
            str, SerializationMode, t.Sequence[t.Union[str, SerializationMode]]
        ],
        *item_cls: t.Type["Item"],
    ):
        if isinstance(smodes, str) or isinstance(smodes, SerializationMode):
            smodes = [smodes]
        self._target_modes = [
            m if isinstance(m, SerializationMode) else SerializationMode[m]
            for m in smodes
        ]
        self._items = set(
            item_cls
            if item_cls
            else ItemFactory.ITEM_DISABLED_SERIALIZATION_MODES.keys()
        )

    def __enter__(self):
        self._prev_modes = {
            itc: ItemFactory.ITEM_DISABLED_SERIALIZATION_MODES[itc]
            for itc in self._items
        }
        set_item_disabled_serialization_modes(self._target_modes, *self._items)

    def __exit__(self, exc_type, exc_value, traceback):
        for itc, val in self._prev_modes.items():
            ItemFactory.set_disabled_serialization_modes(itc, val)


class no_data_cache(ContextDecorator):
    """Use this class as context manager or function decorator to disable data caching
    on some or all item types.

    .. code-block::
       :caption: Example

       # disable data cache for all items
       with no_data_cache():
           ...

       # disable only for BinaryItem and NumpyItem
       with no_data_cache(BinaryItem, NumpyItem):
           ...

       # apply at function invocation
       @no_data_cache(ImageItem)
       def my_fn():
           ...
    """

    def __init__(self, *item_cls: t.Type["Item"]):
        self._items = item_cls if item_cls else ItemFactory.ITEM_DATA_CACHE_MODE.keys()

    def __enter__(self):
        self._prev_state = {
            itc: ItemFactory.ITEM_DATA_CACHE_MODE[itc] for itc in self._items
        }
        disable_item_data_cache(*self._items)

    def __exit__(self, exc_type, exc_value, traceback):
        for itc, val in self._prev_state.items():
            ItemFactory.set_data_cache_mode(itc, val)


class data_cache(ContextDecorator):
    """Use this class as context manager or function decorator to enable data caching
    on some or all item types. Useful when nested with ``no_data_cache``.

    .. code-block::
       :caption: Example

       # disable data cache for all items, then re-enable it
       with no_data_cache(...):
           with data_cache(...):
               ...
           ...
    """

    def __init__(self, *item_cls: t.Type["Item"]):
        self._items = item_cls if item_cls else ItemFactory.ITEM_DATA_CACHE_MODE.keys()

    def __enter__(self):
        self._prev_state = {
            itc: ItemFactory.ITEM_DATA_CACHE_MODE[itc] for itc in self._items
        }
        enable_item_data_cache(*self._items)

    def __exit__(self, exc_type, exc_value, traceback):
        for itc, val in self._prev_state.items():
            ItemFactory.set_data_cache_mode(itc, val)


@contextmanager
def _unclosable_BytesIO():
    """A BytesIO that is closed only when exiting this context."""
    fd = io.BytesIO()
    close = fd.close
    fd.close = lambda: None
    yield fd
    fd.close = close
    fd.close()


T = t.TypeVar("T")
_item_data_source = t.Union[Path, ParseResult]
_item_init_types = t.Union["Item", Path, ParseResult, t.BinaryIO, t.Any]


class Item(t.Generic[T], metaclass=ItemFactory):  # type: ignore
    """Base class for any supported Item. Concrete classes should ideally implement just
    the abstract methods, leaving ``__init__`` as is.
    """

    _data_cache: t.Optional[T]
    _file_sources: t.List[Path]
    _remote_sources: t.List[ParseResult]
    _cache_data: t.Optional[bool]
    _shared: bool
    _serialization_mode: t.Optional[SerializationMode]

    def __init__(self, *sources: _item_init_types, shared: bool = False):
        super().__init__()
        self._data_cache = None
        self._file_sources = []
        self._remote_sources = []
        self._cache_data = None
        self._shared = shared
        self._serialization_mode = None

        # if not sources:
        #     logger.warning(f"{self.__class__}: no source data.")

        for src in sources:
            if isinstance(src, Item):
                src = src()

            if isinstance(src, (Path, ParseResult)):
                self._add_data_source(src)
            elif self._data_cache is not None:
                raise ValueError(f"{self.__class__.__name__}: Cannot set data twice.")
            elif isinstance(src, (t.BinaryIO, io.IOBase)):
                self._data_cache = self.decode(src)
            else:
                self._data_cache = self.validate(src)

    @property
    def cache_data(self) -> t.Optional[bool]:
        return self._cache_data

    @cache_data.setter
    def cache_data(self, enabled: t.Optional[bool]):
        self._cache_data = enabled

    @property
    def serialization_mode(self) -> t.Optional[SerializationMode]:
        return self._serialization_mode

    @serialization_mode.setter
    def serialization_mode(self, mode: t.Optional[t.Union[SerializationMode, str]]):
        self._serialization_mode = (
            mode
            if mode is None or isinstance(mode, SerializationMode)
            else SerializationMode[mode]
        )

    @property
    def is_shared(self) -> bool:
        return self._shared

    def effective_serialization_mode(self) -> SerializationMode:
        smode = (
            self.serialization_mode
            if self.serialization_mode is not None
            else ItemFactory.get_serialization_mode(self.__class__)
        )
        if smode is SerializationMode.REMOTE_FILE and (
            not self._remote_sources
            or not self.is_mode_enabled(SerializationMode.REMOTE_FILE)
        ):
            smode = SerializationMode.HARD_LINK
        return smode

    def is_mode_enabled(self, mode: SerializationMode) -> bool:
        return mode not in ItemFactory.get_disabled_serialization_modes(self.__class__)

    @classmethod
    def make_new(cls, *sources: _item_init_types, shared: bool = False) -> Item:
        return cls(*sources, shared=shared)

    @classmethod
    def as_default_name(cls, filepath: Path) -> Path:
        return filepath.with_suffix(cls.file_extensions()[0])

    @classmethod
    def as_default_remote_file(cls, filepath: Path) -> Path:
        filename = cls.as_default_name(filepath)
        return filename.parent / (filename.name + Item.REMOTE_FILE_EXT)

    @classmethod
    def get_all_names(cls, filepath: Path) -> t.List[Path]:
        names = [filepath.with_suffix(ext) for ext in cls.file_extensions()]
        names += [cls.as_default_remote_file(filepath)]
        return names

    def _check_source(self, source: _item_data_source):
        src_path = Path(source.path) if isinstance(source, ParseResult) else source
        if src_path.suffix not in self.file_extensions():
            srcstr = source.geturl() if isinstance(source, ParseResult) else str(source)
            raise ValueError(
                f"{self.__class__.__name__}: invalid extension for `{srcstr}`"
            )

    def _add_data_source(self, source: _item_data_source) -> bool:
        self._check_source(source)
        if isinstance(source, Path):
            source = source.resolve()
            if source not in self._file_sources:
                self._file_sources.append(source)
                return True
        elif source not in self._remote_sources:
            self._remote_sources.append(source)
            return True
        return False

    def _serialize_to_local_file(self, path: Path) -> t.Optional[Path]:  # noqa: C901
        def _try_copy(copy_fn: t.Callable[[str, str], None], p: str) -> bool:
            for f in self._file_sources:
                try:
                    copy_fn(str(f), p)
                    return True
                except Exception:
                    # logger.exception(f"{self.__class__}: data serialization error.")
                    pass
            return False

        target_path = path.resolve()
        smode: t.Optional[SerializationMode] = self.effective_serialization_mode()

        # At this point if smode is REMOTE_FILE, then REMOTE_FILE is not disabled.
        if smode is SerializationMode.REMOTE_FILE:
            path = self.as_default_remote_file(target_path)
        elif target_path.suffix not in self.file_extensions():
            path = self.as_default_name(target_path)

        # first delete the existing file, if any
        path.unlink(missing_ok=True)

        if smode is SerializationMode.REMOTE_FILE:
            try:
                with path.open("w") as fp:
                    json.dump([rm.geturl() for rm in self._remote_sources], fp)
                return None  # do not save remote file among file sources!
            except Exception as exc:
                logger.warning(  # logger.exception(
                    f"{self.__class__}: remote file serialization error `{exc}`."
                )
                path.unlink(missing_ok=True)
                path = (
                    self.as_default_name(target_path)
                    if target_path.suffix not in self.file_extensions()
                    else target_path
                )
                smode = SerializationMode.HARD_LINK

        if smode is SerializationMode.HARD_LINK:
            smode = (
                None
                if (
                    self.is_mode_enabled(SerializationMode.HARD_LINK)
                    and _try_copy(os.link, str(path))
                )
                else SerializationMode.DEEP_COPY
            )

        if smode is SerializationMode.SYM_LINK:
            smode = (
                None
                if self.is_mode_enabled(SerializationMode.SYM_LINK)
                and _try_copy(os.symlink, str(path))
                else SerializationMode.DEEP_COPY
            )

        if smode is SerializationMode.DEEP_COPY:
            smode = (
                None
                if self.is_mode_enabled(SerializationMode.DEEP_COPY)
                and _try_copy(shutil.copy, str(path))
                else SerializationMode.CREATE_NEW_FILE
            )

        if smode is SerializationMode.CREATE_NEW_FILE and self.is_mode_enabled(
            SerializationMode.CREATE_NEW_FILE
        ):
            data = self()
            if data is not None:
                try:
                    with path.open("wb") as fp:
                        self.encode(data, fp)
                    smode = None
                except Exception as exc:
                    logger.warning(  # logger.exception(
                        f"{self.__class__}: new file serialization error `{exc}`."
                    )
                    pass

        if smode is None:
            return path
        logger.warning(f"{self.__class__}: cannot serialize item data.")
        return None

    def _serialize_to_remote(self, remote_url: ParseResult) -> t.Optional[ParseResult]:
        remote, rm_paths = plr.create_remote(remote_url), plr.paths_from_url(remote_url)
        if remote is not None and rm_paths[0] is not None:
            data = self()
            if data is None:
                logger.warning(f"{self.__class__}: no data to upload to remote.")
            else:
                with _unclosable_BytesIO() as data_stream:
                    self.encode(data, data_stream)

                    data_stream.seek(0, io.SEEK_END)
                    data_size = data_stream.tell()
                    data_stream.seek(0, io.SEEK_SET)

                    return remote.upload_stream(
                        data_stream, data_size, rm_paths[0], self.file_extensions()[0]
                    )
        return None

    def serialize(self, *targets: _item_data_source):
        for trg in targets:
            data_source = (
                self._serialize_to_local_file(trg)
                if isinstance(trg, Path)
                else self._serialize_to_remote(trg)
            )
            if data_source is not None:
                self._add_data_source(data_source)

    def remove_data_source(self, *sources: _item_data_source) -> Item:
        def _normalize_source(src: _item_data_source) -> t.List[_item_data_source]:
            if isinstance(src, Path):
                src = src.resolve()
                # Path.parents supports slicing only from 3.10 onwards
                pp = [src] + [p for p in src.parents]
                pp.pop()  # remove last element, which is `.`, `/`, `c:\` etc.
                return pp  # type: ignore
            src_path = Path(src.path)
            if src_path.suffix:
                src_path = src_path.parent
            return [
                ParseResult(
                    scheme=src.scheme,
                    netloc=src.netloc if src.netloc else "localhost",
                    path=src_path.as_posix(),
                    params="",
                    query="",
                    fragment="",
                )
            ]

        to_be_removed = [pp for src in sources for pp in _normalize_source(src)]

        new_sources: t.List[t.Any] = []
        for src in self._file_sources:
            src_ps = src.parents
            for tbr in to_be_removed:
                if tbr in src_ps:
                    break
            else:
                new_sources.append(src)

        new_sources += [
            src
            for src in self._remote_sources
            if _normalize_source(src)[0] not in to_be_removed
        ]
        new_sources += (
            [self._data_cache]
            if self._data_cache is not None
            else ([self()] if not new_sources else [])
        )

        return self.make_new(*new_sources, shared=self.is_shared)

    def __call__(self) -> t.Optional[T]:
        if self._data_cache is not None:
            return self._data_cache
        for fsrc in self._file_sources:
            try:
                with open(fsrc, "rb") as fp:
                    return self._decode_and_store(fp)
            except Exception as exc:
                logger.warning(  # logger.exception(
                    f"{self.__class__}: file source error `{exc}`."
                )
        for rmsrc in self._remote_sources:
            remote, rm_paths = plr.create_remote(rmsrc), plr.paths_from_url(rmsrc)
            if (
                remote is not None
                and rm_paths[0] is not None
                and rm_paths[1] is not None
            ):
                try:
                    with _unclosable_BytesIO() as data_stream:
                        if remote.download_stream(
                            data_stream, rm_paths[0], rm_paths[1]
                        ):
                            data_stream.seek(0)
                            return self._decode_and_store(data_stream)
                except Exception as exc:
                    logger.warning(  # logger.exception(
                        f"{self.__class__}: remote source error `{exc}`."
                    )
        return None

    def _decode_and_store(self, fp: t.BinaryIO) -> T:
        v = self.decode(fp)
        if (
            self.cache_data is None
            and Item.is_cache_enabled(self.__class__)
            or self.cache_data
        ):
            self._data_cache = v
        return v

    @classmethod
    @abstractmethod
    def file_extensions(cls) -> t.Sequence[str]:
        """Returns the list of valid file extensions with leading dot. The first one is
        assumed to be associated to the encoding applied by `encode`.

        :return: the list of valid file extensions.
        :rtype: t.Sequence[str]
        """
        return []

    @classmethod
    @abstractmethod
    def decode(cls, fp: t.BinaryIO) -> T:
        """Reads data from a binary stream raising an exception if the operation cannot
        complete.

        :param fp: the binary strem to decode.
        :type fp: t.BinaryIO
        :return: the value extracted
        :rtype: T
        """
        pass

    @classmethod
    @abstractmethod
    def encode(cls, value: T, fp: t.BinaryIO) -> None:
        """Writes data to a binary stream raising an exception if the operation cannot
        complete. The encoding should be described by the first extension returned by
        `file_extensions`.

        :param value: the data to write.
        :type value: T
        :param fp: the binary stream.
        :type fp: t.BinaryIO
        """
        pass

    @classmethod
    def validate(cls, raw_data: t.Any) -> T:
        """Subclasses can override to validate or process a raw value before storing.
        Raise an exception if the operation cannot complete.

        :param raw_data: the raw value got from the user.
        :type raw_data: t.Any
        :return: the processed value that can be internally stored.
        :rtype: T
        """
        return raw_data

    def __repr__(self) -> str:
        return (
            f"{self.__class__}(data={repr(self._data_cache)}, "
            f"sources={self._file_sources}, remotes={self._remote_sources}) "
            f"shared={self.is_shared}, cache={self.cache_data}, "
            f"serialization={self.serialization_mode})"
        )

    def __str__(self) -> str:
        str_srcs = [f"    - {str(fs)}" for fs in self._file_sources]
        str_rmts = [f"    - {pr.geturl()}" for pr in self._remote_sources]
        return "\n".join(
            [
                f"{self.__class__.__name__}:",
                f"  data: {str(self._data_cache)}",
                "  sources:",
            ]
            + str_srcs
            + ["  remotes:"]
            + str_rmts
            + [
                f"  shared: {self.is_shared}",
                f"  cache: {self.cache_data}",
                f"  serialization: {self.serialization_mode}",
            ]
        )

    def __pl_pretty__(self) -> t.Any:
        from rich.tree import Tree
        from rich.panel import Panel
        from rich import box

        item_tree = Tree(f"{self.__class__.__name__}")
        if self._data_cache is None:
            item_tree.add("data")
        else:
            item_tree.add(
                Panel.fit(
                    self.pl_pretty_data(self._data_cache),
                    title="data",
                    title_align="left",
                    box=box.HORIZONTALS,
                )
            )
        branch = item_tree.add("sources")
        for fs in self._file_sources:
            branch.add(str(fs))
        branch = item_tree.add("remotes")
        for pr in self._remote_sources:
            branch.add(pr.geturl())
        item_tree.add(f"shared: {self.is_shared}")
        item_tree.add(f"cache: {self.cache_data}")
        item_tree.add(f"serialization: {self.serialization_mode}")
        return item_tree

    @classmethod
    def pl_pretty_data(cls, value: T) -> t.Any:
        return str(value)


class UnknownItem(Item[t.Any]):
    @classmethod
    def file_extensions(cls) -> t.Sequence[str]:
        return ["._"]

    @classmethod
    def decode(cls, fp: t.BinaryIO) -> t.Any:
        raise NotImplementedError(f"{cls.__name__}: cannot decode.")  # pragma: no cover

    @classmethod
    def encode(cls, value: t.Any, fp: t.BinaryIO):
        raise NotImplementedError(f"{cls.__name__}: cannot encode.")  # pragma: no cover



from typing import Any, Dict, List, Union

import pipelime.choixe.ast.nodes as ast
from pipelime.choixe.ast.parser import DIRECTIVE_PREFIX


class Unparser(ast.NodeVisitor):
    """`NodeVisitor` for the `unparse` operation."""

    def visit_dict(self, node: ast.DictNode) -> Dict:
        data = {}
        for k, v in node.nodes.items():
            data[k.accept(self)] = v.accept(self)
        return data

    def visit_list(self, node: ast.ListNode) -> List:
        data = []
        for x in node.nodes:
            data.append(x.accept(self))
        return data

    def visit_literal(self, node: ast.LiteralNode) -> Any:
        return node.data

    def visit_dict_bundle(self, node: ast.DictBundleNode) -> Any:
        data = {}
        for x in node.nodes:
            data.update(x.accept(self))
        return data

    def visit_str_bundle(self, node: ast.StrBundleNode) -> str:
        return "".join(x.accept(self) for x in node.nodes)

    def visit_sweep(self, node: ast.SweepNode) -> Union[Dict, str]:
        return self._unparse_auto("sweep", *node.cases)

    def visit_var(self, node: ast.VarNode) -> Union[Dict, str]:
        kwargs = {}
        if node.default is not None:
            kwargs["default"] = node.default
        if node.env is not None:
            kwargs["env"] = node.env
        if node.help is not None:
            kwargs["help"] = node.help
        return self._unparse_auto("var", node.identifier, **kwargs)

    def visit_import(self, node: ast.ImportNode) -> Union[Dict, str]:
        return self._unparse_auto("import", node.path)

    def visit_symbol(self, node: ast.SymbolNode) -> Any:
        return self._unparse_auto("symbol", node.symbol)

    def visit_instance(self, node: ast.InstanceNode) -> Dict[str, Any]:
        return {
            self._unparse_compact("call"): node.symbol.accept(self),
            self._unparse_compact("args"): node.args.accept(self),
        }

    def visit_model(self, node: ast.ModelNode) -> Dict[str, Any]:
        return {
            self._unparse_compact("model"): node.symbol.accept(self),
            self._unparse_compact("args"): node.args.accept(self),
        }

    def visit_for(self, node: ast.ForNode) -> Dict[str, Any]:
        args = []
        if node.identifier is not None:
            args.append(node.identifier)
        key = self._unparse_call("for", node.iterable, *args)
        value = node.body.accept(self)
        return {key: value}

    def visit_switch(self, node: ast.SwitchNode) -> Dict[str, Any]:
        key = self._unparse_call("switch", node.value)

        value = []
        for case in node.cases:
            entry = {}
            entry[self._unparse_compact("case")] = case[0].accept(self)
            entry[self._unparse_compact("then")] = case[1].accept(self)
            value.append(entry)

        if node.default is not None:
            entry = {}
            entry[self._unparse_compact("default")] = node.default.accept(self)
            value.append(entry)

        return {key: value}

    def visit_item(self, node: ast.ItemNode) -> Union[Dict, str]:
        args = []
        if node.identifier is not None:
            args.append(node.identifier)
        return self._unparse_auto("item", *args)

    def visit_index(self, node: ast.IndexNode) -> Union[Dict, str]:
        args = []
        if node.identifier is not None:
            args.append(node.identifier)
        return self._unparse_auto("index", *args)

    def visit_uuid(self, node: ast.UuidNode) -> str:
        return self._unparse_compact("uuid")

    def visit_date(self, node: ast.DateNode) -> Union[Dict, str]:
        args = []
        if node.format is not None:
            args.append(node.format)
        return self._unparse_auto("date", *args)

    def visit_cmd(self, node: ast.CmdNode) -> Union[Dict, str]:
        return self._unparse_auto("cmd", node.command)

    def visit_tmp_dir(self, node: ast.TmpDirNode) -> Union[Dict, str]:
        args = []
        if node.name is not None:
            args.append(node.name)
        return self._unparse_auto("tmp", *args)

    def _unparse_as_arg(self, node: ast.Node) -> str:
        unparsed = node.accept(self)
        if isinstance(unparsed, str):
            if all([x.isidentifier() for x in unparsed.split(".")]):
                return unparsed
            else:
                return f'"{unparsed}"'
        else:
            return str(unparsed)

    def _unparse_as_args(self, *nodes: ast.Node) -> str:
        return ", ".join([self._unparse_as_arg(x) for x in nodes])

    def _unparse_as_kwargs(self, **nodes: ast.Node) -> str:
        return ", ".join([f"{k}={self._unparse_as_arg(v)}" for k, v in nodes.items()])

    def _unparse_compact(self, name: str) -> str:
        return f"{DIRECTIVE_PREFIX}{name}"

    def _unparse_call(self, name: str, *args: ast.Node, **kwargs: ast.Node) -> str:
        parts = []
        if len(args) > 0:
            parts.append(self._unparse_as_args(*args))
        if len(kwargs) > 0:
            parts.append(self._unparse_as_kwargs(**kwargs))
        return f"{self._unparse_compact(name)}({', '.join(parts)})"

    def _unparse_extended(
        self, name: str, *args: ast.Node, **kwargs: ast.Node
    ) -> Dict[str, Any]:
        return {
            "$directive": name,
            "$args": [x.accept(self) for x in args],
            "$kwargs": {k: v.accept(self) for k, v in kwargs.items()},
        }

    def _unparse_auto(
        self, name: str, *args: ast.Node, **kwargs: ast.Node
    ) -> Union[Dict, str]:
        if len(args) == 0 and len(kwargs) == 0:
            return self._unparse_compact(name)

        all_leafs = True
        for x in list(args) + list(kwargs.values()):
            if not isinstance(x, ast.LiteralNode):
                all_leafs = False
                break

        if all_leafs:
            return self._unparse_call(name, *args, **kwargs)

        return self._unparse_extended(name, *args, **kwargs)


def unparse(node: ast.Node) -> Any:
    """Visitor that undoes the parsing operation.

    If a parser transforms an object into an AST Node, the unparser transforms the AST
    Node back to an object that can be parsed into the same AST Node, as close as
    possible to the original object.

    Of course, since the parser discards some irrelevant information like unnecessary
    white spaces around tokens, the unparser cannot magically bring them back.
    For example::

        unparse(parse('$var( my_id  ,default= 10       '))

    results in::

        '$var(my_id, default=10)'

    Args:
        node (Node): The AST node to unparse.

    Returns:
        Any: The unparsed object.
    """
    unparser = Unparser()
    return node.accept(unparser)



from pathlib import Path

import pytest
from pipelime.choixe.utils.io import dump, load

data = {
    "alice": 10,
    "bob": 20,
    "charlie": ["foo", "bar", {"alpha": 10.0, "beta": 20.0}],
}


@pytest.mark.parametrize(["ext"], [["yml"], ["yaml"], ["json"]])
def test_io(tmp_path: Path, ext: str):
    path = tmp_path / f"config.{ext}"
    dump(data, path)
    loaded = load(path)
    assert loaded == data



from pathlib import Path

import numpy as np

import pipelime.items as pli
import pipelime.sequences as pls

# Let's modify the "mini mnist" dataset by:
# 1. Keeping only the samples with even index.
# 2. Inverting the color of the images.
# 3. Adding a new item called "color" with the average image color.
# 4. Deleting the "maskinv" item.

# Create a sequence from an underfolder
this_folder = Path(__file__).parent
seq = pls.SamplesSequence.from_underfolder(this_folder / "datasets/mini_mnist")

# Keep only the even samples (POINT 1)
even_samples = seq[::2]

# Initialize an empty list of samples
new_samples = []

# Iterate on the sequence
for sample in even_samples:

    # Get the image item
    image: np.ndarray = sample["image"]()  # type: ignore

    # Invert the colors
    invimage = 255 - image

    # Replace the value of "image" with the inverted image (POINT 2)
    sample = sample.set_value("image", invimage)

    # Get the average image color
    avg_color = np.mean(invimage, axis=(0, 1))

    # Create a numpy item with the average color and add it to the sample (POINT 3)
    avg_color_item = pli.NpyNumpyItem(avg_color)
    sample = sample.set_item("avg_color", avg_color_item)

    # Delete the maskinv item (POINT 4)
    sample = sample.remove_keys("maskinv")

    # After the sample has been modified, add it to the sequence
    new_samples.append(sample)

# Create a new sequence from the list of samples
new_seq = pls.SamplesSequence.from_list(new_samples)

# Save the new sequence to an underfolder
new_seq = new_seq.to_underfolder(this_folder / "datasets/mini_mnist_inv")
new_seq.run()



import base64
from io import BytesIO
from pathlib import Path
from typing import Optional, Sequence

import numpy as np
import requests
from PIL import Image

from pipelime.piper.drawing.base import NodesGraphDrawer
from pipelime.piper.graph import (
    DAGNodesGraph,
    GraphNode,
    GraphNodeData,
    GraphNodeOperation,
)


class MermaidNodesGraphDrawer(NodesGraphDrawer):
    PRIVATE_CHARS = ["@"]
    SUBSTITUTION_CHAR = "_"
    OPERATION_NAME = "operation"
    DATA_NAME = "data"
    MERMAID_SERVICE_URL = "https://mermaid.ink/img/{}"

    def _is_node_file(self, node: GraphNodeData) -> bool:
        """Returns True if the node path is (or could be) a file

        Args:
            node (GraphNodeData):  input node

        Returns:
            bool:  True if the node path is (or could be) a file
        """
        return len(Path(node.path).suffix) > 0

    def _data_node_shape(self, node: GraphNodeData) -> str:
        """Returns the shape string of a data node

        Args:
            node (GraphNodeData): input node

        Returns:
            str: shape of the node
        """

        return (
            f"{node.name}[[{node.name}]]"
            if self._is_node_file(node)
            else f"{node.name}[({node.name})]"
        )

    def _node_representation(self, node: GraphNode) -> str:
        """Returns the string mermaid representation of a node

        Args:
            node (GraphNode): node to represent

        Raises:
            ValueError: if node is not a valid node

        Returns:
            str: mermaid textual representation of the node
        """

        out = ""
        if isinstance(node, GraphNodeOperation):
            out = f"{node.name}:::{MermaidNodesGraphDrawer.OPERATION_NAME}"
        elif isinstance(node, GraphNodeData):
            out = f"{self._data_node_shape(node)}:::{MermaidNodesGraphDrawer.DATA_NAME}"
        else:
            raise ValueError(f"Unknown node type: {type(node)}")

        for c in MermaidNodesGraphDrawer.PRIVATE_CHARS:
            out = out.replace(c, MermaidNodesGraphDrawer.SUBSTITUTION_CHAR)
        return out

    def _style_attrs_operation(self) -> dict:
        """
        Returns the style attributes for an operation node
        """
        return {"fill": "#03A9F4", "stroke": "#fafafa", "color": "#fff"}

    def _style_attrs_data(self) -> dict:
        """
        Returns the style attributes for a data node
        """
        return {"fill": "#009688", "stroke": "#fafafa", "color": "#fff"}

    def _classdef_row(self, name: str, style_attrs: dict) -> str:
        """Returns the string representation of a class (style) definition

        Args:
            name (str): name of the class
            style_attrs (dict): style attributes of the class

        Returns:
            str: string representation of the class definition
        """

        def style_string(x):
            return ",".join([f"{k}:{v}" for k, v in x.items()])

        return f"classDef {name} {style_string(style_attrs)};"

    def _header_row(self, orientation: str = "TB") -> str:
        """Returns the string representation of the header row of the mermaid graph

        Args:
            orientation (str, optional): TB - top to bottom, TD - top-down/ same as top
            to bottom, BT - bottom to top, RL - right to left, LR - left to right .
            Defaults to "TB".

        Returns:
            str: string representation of the header row
        """
        return f"flowchart {orientation}"

    def _edge_row(self, graph: DAGNodesGraph, n0: GraphNode, n1: GraphNode) -> str:
        """Returns the string representation of an edge row in the mermaid graph

        Args:
            graph (DAGNodesGraph): graph to draw
            n0 (GraphNode): first node of the edge
            n1 (GraphNode): second node of the edge

        Returns:
            str: string representation of the edge row
        """

        edge_attrs = graph.raw_graph.edges[n0, n1]

        edge_name = ""
        if DAGNodesGraph.GraphAttrs.INPUT_PORT in edge_attrs:
            edge_name = f"|{edge_attrs[DAGNodesGraph.GraphAttrs.INPUT_PORT]}|"
        if DAGNodesGraph.GraphAttrs.OUTPUT_PORT in edge_attrs:
            edge_name += f"|{edge_attrs[DAGNodesGraph.GraphAttrs.OUTPUT_PORT]}|"

        return "{}-->{}{}".format(
            self._node_representation(n0),
            edge_name,
            self._node_representation(n1),
        )

    def _add_row(self, representation: str, row: str, header: bool = False):
        """Append a row to the representation with correct indentation and newline

        Args:
            representation (str): representation to append to
            row (str): row to append
            header (bool, optional): if TRUE the row is a header row. Defaults to False.

        Returns:
            _type_: representation with the row appended
        """
        representation = representation + (f"{row}\n" if header else f"\t{row}\n")
        return representation

    def draw(self, graph: DAGNodesGraph) -> np.ndarray:
        """Draws a graph and returns the image as a numpy array

        Args:
            graph (DAGNodesGraph): graph to draw

        Returns:
            np.ndarray: image as a numpy array
        """

        representation = self.representation(graph)
        graphbytes = representation.encode("ascii")
        base64_bytes = base64.b64encode(graphbytes)
        base64_string = base64_bytes.decode("ascii")
        url = MermaidNodesGraphDrawer.MERMAID_SERVICE_URL.format(base64_string)
        response = requests.get(url)
        return np.array(Image.open(BytesIO(response.content)))

    def representation(self, graph: DAGNodesGraph) -> str:
        """Returns the string representation of the graph in the mermaid language

        Args:
            graph (DAGNodesGraph): graph to draw

        Returns:
            str: string representation of the graph in the mermaid language
        """

        # execution_stack = graph.build_execution_stack()

        # initialize representation
        repr = ""

        # flochart header
        repr = self._add_row(repr, self._header_row(), header=True)

        # style calss for operations
        repr = self._add_row(
            repr,
            self._classdef_row(self.OPERATION_NAME, self._style_attrs_operation()),
        )

        # style class for data
        repr = self._add_row(
            repr,
            self._classdef_row(self.DATA_NAME, self._style_attrs_data()),
        )

        for edge in graph.raw_graph.edges():
            n0: GraphNode = edge[0]
            n1: GraphNode = edge[1]

            repr = self._add_row(repr, self._edge_row(graph, n0, n1))

        return repr

    def exportable_formats(self) -> Sequence[str]:
        """Returns the exportable formats of the graph drawer.

        Returns:
            Sequence[str]: exportable formats of the graph drawer
        """
        return ["png", "markdown", "md"]

    def export(self, graph: DAGNodesGraph, filename: str, format: Optional[str] = None):
        super().export(graph, filename, format)

        if format is None:
            format = Path(filename).suffix[1:]

        if format == "png":
            import imageio

            img = self.draw(graph)
            imageio.imwrite(filename, img)
        elif format in ["markdown", "md"]:
            representation = self.representation(graph)
            markdown = f"\n```mermaid\n{representation}\n```\n"
            with open(filename, "w") as f:
                f.write(markdown)



class TestCliBase:
    def _base_launch(self, args: list):
        from typer.testing import CliRunner

        from pipelime.cli.main import app

        runner = CliRunner()
        result = runner.invoke(app, args)
        print(result.output)
        assert result.exit_code == 0
        assert result.exception is None
        return result

    def test_help(self, extra_modules):
        for module_data in extra_modules:
            for cmd in module_data["operators"] + module_data["commands"]:
                result = self._base_launch(
                    ["-m", str(module_data["filepath"]), "help", cmd]
                )
                assert cmd in result.output

    def test_list(self, extra_modules):
        def _check(result, *, in_modules=[], not_in_modules=[]):
            for v in in_modules:
                assert v in result.output
            for v in not_in_modules:
                assert v not in result.output

        args = [
            v
            for module_data in extra_modules
            for v in ("-m", str(module_data["filepath"]))
        ] + ["list"]

        result = self._base_launch(args)
        for module_data in extra_modules:
            _check(
                result, in_modules=module_data["operators"] + module_data["commands"]
            )

        result = self._base_launch(["--verbose"] + args)
        for module_data in extra_modules:
            _check(
                result, in_modules=module_data["operators"] + module_data["commands"]
            )
        assert "Fields" in result.output
        assert "Description" in result.output
        assert "Type" in result.output
        assert "Default" in result.output

    def test_run(self, data_folder, minimnist_dataset, tmp_path):
        from pathlib import Path
        from typing import Sequence

        import numpy as np

        from pipelime.items import NpyNumpyItem
        from pipelime.sequences import SamplesSequence

        outpath = tmp_path / "output"
        args = [
            "pipe",
            "+input.folder",
            str(minimnist_dataset["path"]),
            "+output.folder",
            str(outpath),
            "+output.serialization.override.DEEP_COPY",
            "NpyNumpyItem",
            "+operations.slice.stop",
            "10",
        ]

        self._base_launch(args)

        outreader = SamplesSequence.from_underfolder(outpath)  # type: ignore
        gt = (
            SamplesSequence.from_underfolder(minimnist_dataset["path"])  # type: ignore
            .slice(stop=10)
        )
        assert len(outreader) == len(gt)
        for o, g in zip(outreader, gt):
            assert o.keys() == g.keys()
            for k, v in o.items():
                if isinstance(v(), np.ndarray):
                    assert np.array_equal(v(), g[k](), equal_nan=True)  # type: ignore
                else:
                    assert v() == g[k]()
                assert isinstance(v._file_sources, Sequence)
                assert len(v._file_sources) == 1
                path = Path(v._file_sources[0])
                assert not path.is_symlink()
                assert path.is_file()
                assert (
                    path.stat().st_nlink == 1
                    if isinstance(v, NpyNumpyItem)
                    else path.stat().st_nlink > 1
                )



import typing as t

import pydantic as pyd

import pipelime.sequences as pls
from pipelime.sequences.pipes import PipedSequenceBase


@pls.piped_sequence
class ValidatedSequence(
    PipedSequenceBase, title="validate_samples", arbitrary_types_allowed=True
):
    """Validates the source sequence against a schema."""

    sample_schema: t.Type[pyd.BaseModel] = pyd.Field(
        ...,
        description=(
            "The sample schema to validate, ie, a pydantic model where fields name are "
            "the sample keys and field types are the expected item classes."
        ),
    )
    lazy: bool = pyd.Field(
        False, description="If True, samples will be validated only when accessed."
    )
    max_samples: int = pyd.Field(
        1,
        description=(
            "When the validation is NOT lazy, "
            "only the slice `[0:max_samples]` is checked. "
            "Set to 0 to check all the samples."
        ),
    )

    def __init__(self, **data):
        super().__init__(**data)
        if not self.lazy:
            seq = (
                self.source
                if self.max_samples == 0
                else self.source[0 : self.max_samples]  # noqa
            )
            for sample in seq:  # noqa: E203
                self._check_sample(sample)  # type: ignore

    def get_sample(self, idx: int) -> pls.Sample:
        sample = self.source[idx]
        if self.lazy:
            self._check_sample(sample)
        return sample

    def _check_sample(self, sample: pls.Sample):
        from pydantic import ValidationError
        from pydantic.error_wrappers import display_errors

        try:
            _ = self.sample_schema(**sample)  # type: ignore
        except ValidationError as e:
            errs = e.errors()
            raise ValueError(
                f"Sample schema validation failed for:\n{str(sample)}\n\n"
                f"Errors:\n{display_errors(errs)}"
            ) from e



from pipelime.piper.progress.listener.base import ListenerCallback, ProgressReceiver
from pipelime.piper.progress.listener.receiver.zmq import ZMQProgressReceiver
from pipelime.piper.progress.listener.callbacks.rich_table import (
    RichTableListenerCallback,
)


class ProgressReceiverFactory:
    """Factory for ``ProgressReceiver`` s"""

    DEFAULT_RECEIVER_TYPE = "ZMQ"

    CLASS_MAP = {
        "ZMQ": ZMQProgressReceiver,
    }

    @classmethod
    def get_receiver(
        cls, token: str, type_: str = DEFAULT_RECEIVER_TYPE
    ) -> ProgressReceiver:
        return cls.CLASS_MAP[type_](token)


class ListenerCallbackFactory:
    """Factory for ``ListenerCallback`` s"""

    DEFAULT_CALLBACK_TYPE = "RICH_TABLE"

    CLASS_MAP = {
        "RICH_TABLE": RichTableListenerCallback,
    }

    @classmethod
    def get_callback(
        cls, type_: str = DEFAULT_CALLBACK_TYPE, **kwargs
    ) -> ListenerCallback:
        return cls.CLASS_MAP[type_](**kwargs)



import typing as t

import pydantic as pyd

import pipelime.sequences as pls


@pls.source_sequence
class SamplesList(pls.SamplesSequence, title="from_list", arbitrary_types_allowed=True):
    """A SamplesSequence from a list of Samples."""

    samples: t.Sequence[pls.Sample] = pyd.Field(
        ..., description="The source sequence of samples."
    )

    def __init__(self, samples: t.Sequence[pls.Sample], **data):
        super().__init__(samples=samples, **data)  # type: ignore

    def size(self) -> int:
        return len(self.samples)

    def get_sample(self, idx: int) -> pls.Sample:
        return self.samples[idx]



from typing import List
import pytest
from pipelime.piper.progress.tracker.base import TrackCallback, Tracker
from pipelime.piper.progress.model import ProgressUpdate, OperationInfo


class MockTrackCallback(TrackCallback):
    def __init__(self):
        super().__init__()
        self.on_start_calls: List[ProgressUpdate] = []
        self.on_advance_calls: List[ProgressUpdate] = []
        self.on_finish_calls: List[ProgressUpdate] = []

    def on_start(self, prog: ProgressUpdate) -> None:
        self.on_start_calls.append(prog)

    def on_advance(self, prog: ProgressUpdate) -> None:
        self.on_advance_calls.append(prog)

    def on_finish(self, prog: ProgressUpdate) -> None:
        self.on_finish_calls.append(prog)


class TestTrackCallback:
    def test_track_callback(self):
        # Create callback
        N = 10
        callback = MockTrackCallback()

        # Assert that callback raises an error if not setup
        with pytest.raises(RuntimeError):
            callback.advance()
        with pytest.raises(RuntimeError):
            callback.finish()

        # Setup callback
        op_info = OperationInfo(
            token="token", node="node", chunk=1, message="message", total=N
        )
        callback.start(op_info)

        # Assert that callback raises an error if already setup
        with pytest.raises(RuntimeError):
            callback.start(op_info)

        # Advance callback
        for _ in range(N // 2):
            callback.advance(2)

        # Finish callback
        callback.finish()

        # Assert that callback was called correctly
        assert len(callback.on_start_calls) == 1
        assert len(callback.on_advance_calls) == N // 2
        assert len(callback.on_finish_calls) == 1

        assert callback.on_advance_calls[-1].progress == N


class TestTracker:
    def test_tracker(self):
        # Create tracker
        callback = MockTrackCallback()
        tracker = Tracker("token", "node", callback)

        # Create a list of stuff
        seq = list(range(10, 20, 2))

        # Track the list
        decorated = tracker.track(seq)

        # Assert that the list was not modified by tracking
        assert list(decorated) == seq

        # Assert that callback was called correctly
        assert len(callback.on_start_calls) == 1
        assert len(callback.on_advance_calls) == len(seq)
        assert len(callback.on_finish_calls) == 1

        assert callback.on_advance_calls[-1].progress == len(seq)



from pipelime.piper.progress.estimator.factory import EstimatorFactory
from pipelime.piper.progress.estimator.base import Estimator


class TestEstimatorFactory:
    def test_get_estimator(self):
        estimator = EstimatorFactory.get_estimator()
        assert isinstance(estimator, Estimator)



from pipelime.piper.parsers.factory import DAGParserFactory
from pipelime.piper.parsers.base import DAGParser


class TestDAGParserFactory:
    def test_available_parsers(self):
        # Get available parsers
        available = DAGParserFactory.available_parsers()

        # Check if the list contains at least one parser
        assert isinstance(available, list)
        assert len(available) > 0

        # Check that the available parsers are all strings
        for x in available:
            assert isinstance(x, str)

    def test_get_parser(self):
        # Get a parser
        parser = DAGParserFactory.get_parser()

        # Check that the parser is a DAGParser
        assert isinstance(parser, DAGParser)



import importlib
import importlib.util
import os
import sys
from contextlib import ContextDecorator
from pathlib import Path
from types import ModuleType
from typing import Any, Optional, Union
from uuid import uuid1


class add_to_sys_path(ContextDecorator):
    """add_to_sys_path context decorator temporarily adds a path to sys.path"""

    def __init__(self, path: Path) -> None:
        self._new_cwd = path

    def __enter__(self) -> None:
        sys.path.insert(0, str(self._new_cwd))

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        sys.path.pop(0)


class add_to_sys_module(ContextDecorator):
    """
    ☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️
    THIS IS A HACK TO ALLOW DYNAMIC IMPORT FROM FILES CONTAINING DATACLASSES.
    THE USAGE OF THIS DECORATOR SHOULD BE AVOIDED AS IT COULD POTENTIALLY BREAK STUFF.

    THIS DECORATOR IS CURRENTLY USED ONLY IN THIS MODULE AND WE PLAN TO REMOVE IT AS
    SOON AS A BETTER SOLUTION IS FOUND.
    ☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️

    add_to_sys_module context decorator temporarily adds a module to sys.modules
    """

    def __init__(self, module: ModuleType, id_: str) -> None:
        self._module = module
        self._id_ = id_

    def __enter__(self) -> None:
        if self._id_ in sys.modules:  # pragma: no cover
            raise RuntimeError(f"Attempted to overwrite system module {self._id_}")

        sys.modules[self._id_] = self._module

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        sys.modules.pop(self._id_)


def import_module_from_file(
    module_file_path: Union[str, Path], cwd: Optional[Path] = None
) -> ModuleType:
    module_path = Path(module_file_path)
    if not module_path.is_absolute():
        cwd = Path(os.getcwd()) if cwd is None else cwd
        module_path = cwd / module_path

    with add_to_sys_path(module_path.parent):
        id_ = uuid1().hex
        spec = importlib.util.spec_from_file_location(id_, str(module_path))
        if spec is None or spec.loader is None:
            raise ImportError(f"Cannot load module `{module_path}`")  # pragma: no cover
        module_ = importlib.util.module_from_spec(spec)
        with add_to_sys_module(module_, id_):
            spec.loader.exec_module(module_)
        return module_


def import_module_from_path(module_class_path: str) -> ModuleType:
    return importlib.import_module(module_class_path)


def import_module(
    module_file_or_class_path: str, cwd: Optional[Path] = None
) -> ModuleType:
    return (
        import_module_from_file(module_file_or_class_path, cwd)
        if module_file_or_class_path.endswith(".py")
        else import_module_from_path(module_file_or_class_path)
    )


def import_symbol(symbol_path: str, cwd: Optional[Path] = None) -> Any:
    """Dynamically imports a given symbol. A symbol can be either:

    - A filesystem path followed by the name of an object to import, like:
        "path/to/my_file.py:MyClass"

    - A python module path, like "module.submodule.MyClass"

    Anything can be imported: modules, classes, functions, etc...

    Args:
        symbol_path (str): The symbol to import
        cwd (Optional[Path], optional): The current working directory to resolve
        relative imports. If None, the system cwd will be used. Defaults to None.

    Raises:
        ImportError: If anything goes wrong with the import.

    Returns:
        Any: The dynamically imported object.
    """
    try:
        if ":" in symbol_path:
            module_path, _, symbol_name = symbol_path.rpartition(":")
            module_ = import_module_from_file(module_path, cwd)
        else:
            module_path, _, symbol_name = symbol_path.rpartition(".")
            module_ = import_module_from_path(module_path)
        return getattr(module_, symbol_name)
    except Exception as e:
        raise ImportError(f"Cannot import {symbol_path}") from e



from pathlib import Path
from typing import Sequence

from pydantic import Field

from pipelime.piper.model import PipelimeCommand, PiperPortType


class MyUselessCommand(PipelimeCommand):
    input_folder: Path = Field(..., piper_port=PiperPortType.INPUT)
    output_folder: Path = Field(..., piper_port=PiperPortType.OUTPUT)
    n: int = 10

    def run(self) -> None:
        import time

        print(self.input_folder, self.output_folder)
        for x in self.track(range(self.n), message="Doing stuff..."):
            time.sleep(0.5)

        for x in self.track(range(self.n), message="Doing stuff 2..."):
            time.sleep(0.5)


class FakeSum(PipelimeCommand):
    input_folders: Sequence[Path] = Field(..., piper_port=PiperPortType.INPUT)
    output_folder: Path = Field(..., piper_port=PiperPortType.OUTPUT)

    def run(self) -> None:
        import time

        print(self.input_folders, self.output_folder)
        for x in self.track(range(5), message="Summing stuff..."):
            time.sleep(0.5)


if __name__ == "__main__":
    a = MyUselessCommand(
        input_folder=Path("/tmp/input"), output_folder=Path("/tmp/output")
    )
    a.set_piper_info(token="TOKEN", node="UselessCommand_A")
    a()
    a = MyUselessCommand(
        input_folder=Path("/tmp/input"), output_folder=Path("/tmp/output")
    )
    a.set_piper_info(token="TOKEN", node="UselessCommand_B")
    a()
    a = MyUselessCommand(
        input_folder=Path("/tmp/input"), output_folder=Path("/tmp/output")
    )
    a.set_piper_info(token="TOKEN", node="UselessCommand_C")
    a()



import imageio.v3 as iio
import tifffile
import numpy as np
import typing as t

from pipelime.items.numpy_item import NumpyItem


class ImageItem(NumpyItem):
    """Base class for all image types. Subclasses should implement `file_extensions`
    and, optionally, `save_options`.
    """

    @classmethod
    def save_options(cls) -> t.Mapping[str, t.Any]:
        """Subclass can override and return specific saving options.

        :return: saving parameters
        :rtype: t.Mapping[str, t.Any]
        """
        return {}

    @classmethod
    def decode(cls, fp: t.BinaryIO) -> np.ndarray:
        return np.array(iio.imread(fp, extension=cls.file_extensions()[0]))

    @classmethod
    def encode(cls, value: np.ndarray, fp: t.BinaryIO):
        iio.imwrite(fp, value, extension=cls.file_extensions()[0], **cls.save_options())

    @classmethod
    def pl_pretty_data(cls, value: np.ndarray) -> t.Any:
        IMAGE_SIZE = 32
        if value.shape[0] > IMAGE_SIZE or value.shape[1] > IMAGE_SIZE:
            import albumentations as A

            if value.dtype != np.uint8:
                src_min, src_max = value.min(), value.max()
                value = (value.astype(np.float32) + src_min) / float(src_max - src_min)
                value = (value * 255.0).astype(np.uint8)

            value = A.LongestMaxSize(max_size=IMAGE_SIZE, always_apply=True)(
                image=value
            )["image"]

        def _get_color_str(v) -> str:
            return (
                f"{v[0]},{v[1]},{v[2]}" if isinstance(v, np.ndarray) else f"{v},{v},{v}"
            )

        return "\n".join(
            [
                "".join(
                    [
                        f"[rgb({_get_color_str(value[v, u])})]█[/]"
                        for u in range(value.shape[1])
                    ]
                )
                for v in range(value.shape[0])
            ]
        )


class BmpImageItem(ImageItem):
    @classmethod
    def file_extensions(cls) -> t.Sequence[str]:
        return (".bmp",)


class PngImageItem(ImageItem):
    @classmethod
    def save_options(cls) -> t.Mapping[str, t.Any]:
        return {"compress_level": 4}

    @classmethod
    def file_extensions(cls) -> t.Sequence[str]:
        return (".png",)


class JpegImageItem(ImageItem):
    @classmethod
    def save_options(cls) -> t.Mapping[str, t.Any]:
        return {"quality ": 80}

    @classmethod
    def file_extensions(cls) -> t.Sequence[str]:
        return (".jpeg", ".jpg", ".jfif", ".jpe")


class TiffImageItem(ImageItem):
    @classmethod
    def file_extensions(cls) -> t.Sequence[str]:
        return (".tiff", ".tif")

    @classmethod
    def save_options(cls) -> t.Mapping[str, t.Any]:
        return {"compression": "zlib"}

    @classmethod
    def decode(cls, fp: t.BinaryIO) -> np.ndarray:
        return np.array(tifffile.imread(fp))

    @classmethod
    def encode(cls, value: np.ndarray, fp: t.BinaryIO):
        tifffile.imwrite(fp, value, **cls.save_options())



import pytest
from pathlib import Path
import numpy as np
import pipelime.sequences as pls
import pipelime.items as pli


@pytest.mark.parametrize(
    ["num_workers", "keep_order", "prefetch"],
    [
        (0, False, 2),
        (1, False, 2),
        (4, False, 4),
        (4, True, 4),
        (-1, False, 4),
    ],
)
def test_grabber(
    num_workers: int,
    keep_order: bool,
    prefetch: int,
    minimnist_dataset: dict,
    tmp_path: Path,
):
    from copy import deepcopy

    source = pls.SamplesSequence.from_underfolder(  # type: ignore
        folder=minimnist_dataset["path"], merge_root_items=True
    )
    proc = (
        source.slice(stop=10)
        .repeat(10)
        .enumerate(idx_key="ii")
        .to_underfolder(folder=tmp_path / "output")
    )
    grabber = pls.Grabber(
        num_workers=num_workers, keep_order=keep_order, prefetch=prefetch
    )

    itm_sm = pli.item_serialization_mode(
        pli.SerializationMode.CREATE_NEW_FILE,
    )
    pls.grab_all(
        grabber,
        proc,
        grab_context_manager=deepcopy(itm_sm),
        worker_init_fn=itm_sm.__enter__,
    )

    for f in tmp_path.glob("output/**/*"):
        assert not f.is_symlink()
        assert not f.is_file() or f.stat().st_nlink == 1

    dest = pls.SamplesSequence.from_underfolder(  # type: ignore
        folder=tmp_path / "output", merge_root_items=True
    )

    for idx, dst_smpl in enumerate(dest):
        src_smpl = source[idx % 10]
        assert set(src_smpl.keys()) | {"ii"} == set(dst_smpl.keys())
        for k, v1 in src_smpl.items():
            v2 = dst_smpl[k]
            assert v1.__class__ == v2.__class__
            if isinstance(v1, pli.NumpyItem):
                assert np.array_equal(v1(), v2(), equal_nan=True)  # type: ignore
            else:
                assert v1() == v2()



from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union

import pydash as py_
from box import Box
from schema import Schema

from pipelime.choixe.ast.nodes import Node
from pipelime.choixe.ast.parser import parse
from pipelime.choixe.utils.io import dump, load
from pipelime.choixe.utils.common import deep_set_
from pipelime.choixe.visitors import Inspection, decode, inspect, process, walk


class XConfig(Box):
    """A configuration with superpowers!"""

    PRIVATE_KEYS = ["_cwd", "_schema"]
    """Keys to exclude in visiting operations"""

    def __init__(
        self,
        data: Optional[Mapping] = None,
        cwd: Optional[Path] = None,
        schema: Optional[Schema] = None,
    ):
        """Constructor for `XConfig`

        Args:
            data (Optional[Mapping], optional): Optional dictionary containing
            initial data. Defaults to None.
            cwd (Optional[Path], optional): An optional path with the current working
            directory to use when resolving relative imports. If set to None, the
            system current working directory will be used. Defaults to None.
            schema (Optional[Schema], optional): Python schema object used for
            validation. Defaults to None.
        """

        # options
        self._cwd = cwd
        self._schema = None

        data = data if data is not None else {}
        assert isinstance(
            data, Mapping
        ), f'XConfig can only contain mappings, found "{data.__class__.__name__}"'

        self.update(data)
        self.set_schema(schema)

    def __repr__(self) -> str:  # pragma: no cover
        return f"<{self.__class__.__name__}: {self.to_dict()}>"

    @classmethod
    def from_file(
        cls, path: Union[str, Path], schema: Optional[Schema] = None
    ) -> XConfig:
        """Factory method to create a `XConfig` from file.

        Args:
            path (Union[str, Path]): Path to a markup file from which to load the data
            schema (Optional[Schema], optional): Python schema object used for
            validation. Defaults to None.

        Returns:
            XConfig: The loaded `XConfig`
        """
        path = Path(path)
        return XConfig(data=load(path), cwd=path.parent, schema=schema)

    def get_schema(self) -> Optional[Schema]:
        """Getter for the configuration schema"""
        return self._schema

    def set_schema(self, s: Optional[Schema]) -> None:
        """Setter for the configuration schema"""
        assert s is None or isinstance(
            s, Schema
        ), "schema is not a valid Schema object!"
        self._schema = s

    def get_cwd(self) -> Optional[Path]:
        """Getter for the configuration cwd"""
        return self._cwd

    def copy(self) -> XConfig:
        """Prototype method to copy this `XConfig` object.

        Returns:
            XConfig: A deepcopy of this `XConfig`.
        """
        return XConfig(data=self.decode(), cwd=self.get_cwd(), schema=self.get_schema())

    def validate(self, replace: bool = True):
        """Validate internal schema if any

        Args:
            replace (bool, optional): True to replace internal dictionary with
            force-validated fields (e.g. Schema.Use). Defaults to True.
        """

        schema = self.get_schema()
        if schema is not None:
            new_dict = schema.validate(self.to_dict())
            if replace:
                self.update(new_dict)

    def is_valid(self) -> bool:
        """Check for schema validity

        Returns:
            bool: True for valid or no schema inside
        """
        schema = self.get_schema()
        if schema is not None:
            return schema.is_valid(self.to_dict())
        return True

    def save_to(self, filename: Union[str, Path]) -> None:
        """Save configuration to output file

        Args:
            filename (str): output filename
        """
        dump(self.decode(), Path(filename))

    def deep_get(
        self, full_key: Union[str, list], default: Optional[Any] = None
    ) -> Any:
        """Gets value based on full path key, ie, 'a.b.0.d' or ['a','b','0','d']

        Args:
            full_key (Union[str, list]): full path key in pydash notation.
            default (Optional[Any], optional): result in case the path is not present.
            Defaults to None.

        Returns:
            Any: The value at the specified path.
        """
        return py_.get(self, full_key, default=default)

    def deep_set(
        self, full_key: Union[str, list], value: Any, only_valid_keys: bool = True
    ) -> None:
        """Sets value based on full path key, ie, 'a.b.0.d' or ['a','b','0','d'])

        Args:
            full_key (Union[str, list]): Full path key in pydash notation.
            value (Any): The value to set.
            only_valid_keys (bool, optional): True to avoid set on not present keys.
            Defaults to True.
        """

        if not only_valid_keys or py_.has(self, full_key):
            deep_set_(self, key_path=full_key, value=value, append=False)

    def deep_update(self, data: Dict, full_merge: bool = False):
        """Updates current confing in depth, based on keys of other input dictionary.
        It is used to replace nested keys with new ones, but can also be used as a merge
        of two completely different XConfig if ``full_merge=True``.

        Args:
            data (Dict): An other dictionary to use as data source.
            full_merge (bool, optional): False to replace only the keys that are
            actually present. Defaults to False.
        """

        other_chunks = walk(parse(data))
        for key, new_value in other_chunks:
            self.deep_set(key, new_value, only_valid_keys=not full_merge)

    def parse(self) -> Node:
        """Parse this object into a Choixe AST Node.

        Returns:
            Node: The parsed node.
        """
        sanitized = dict(self)
        [sanitized.pop(x) for x in self.PRIVATE_KEYS]
        return parse(sanitized)

    def decode(self) -> Dict:
        """Convert this XConfig to a plain python dictionary. Also converts some nodes
        like numpy arrays into plain lists and restores some directives, eg, `$model`.

        Returns:
            Dict: The decoded dictionary.
        """
        return decode(self.parse())

    def to_dict(self) -> Dict:
        """Convert this XConfig to a plain python dictionary with no value decoding.

        Returns:
            Dict: The plain dictionary.
        """
        return {
            k: v for k, v in super().to_dict().items() if k not in self.PRIVATE_KEYS
        }

    def walk(self) -> List[Tuple[List[Union[str, int]], Any]]:
        """Perform the walk operation on this XConfig.

        Returns:
            List[Tuple[List[Union[str, int]], Any]]: The walk output.
        """
        return walk(self.parse())

    def _process(
        self, context: Optional[Dict[str, Any]] = None, allow_branching: bool = True
    ) -> List[XConfig]:
        data = process(
            self.parse(),
            context=context,
            cwd=self.get_cwd(),
            allow_branching=allow_branching,
        )
        return [
            XConfig(data=x, cwd=self.get_cwd(), schema=self.get_schema()) for x in data
        ]

    def process(self, context: Optional[Dict[str, Any]] = None) -> XConfig:
        """Process this XConfig without branching.

        Args:
            context (Optional[Dict[str, Any]], optional): Optional data structure
            containing all variables values. Defaults to None.

        Returns:
            XConfig: The processed XConfig.
        """
        return self._process(context=context, allow_branching=False)[0]

    def process_all(self, context: Optional[Dict[str, Any]] = None) -> List[XConfig]:
        """Process this XConfig with branching.

        Args:
            context (Optional[Dict[str, Any]], optional): Optional data structure
            containing all variables values. Defaults to None.

        Returns:
            List[XConfig]: A list of all processing outcomes.
        """
        return self._process(context=context, allow_branching=True)

    def inspect(self) -> Inspection:
        return inspect(self.parse(), cwd=self.get_cwd())
